import asyncio
from copy import copy

from sqlalchemy.engine import url
from sqlalchemy import util
from sqlalchemy.engine.strategies import EngineStrategy

from .engine import GinoEngine


class GinoStrategy(EngineStrategy):
    """A SQLAlchemy engine strategy for GINO.

    This strategy is initialized automatically as :mod:`gino` is imported.

    If :func:`sqlalchemy.create_engine` uses ``strategy="gino"``, it will return a
    :class:`~collections.abc.Coroutine`, and treat URL prefix ``postgresql://`` or
    ``postgres://`` as ``postgresql+asyncpg://``.
    """

    name = "gino"
    engine_cls = GinoEngine

    async def create(self, name_or_url, loop=None, **kwargs):
        engine_cls = self.engine_cls
        u = url.make_url(name_or_url)
        if loop is None:
            loop = asyncio.get_event_loop()
        if u.drivername in {"postgresql", "postgres"}:
            u = copy(u)
            u.drivername = "postgresql+asyncpg"

        dialect_cls = u.get_dialect()

        pop_kwarg = kwargs.pop

        dialect_args = {}
        # consume dialect arguments from kwargs
        for k in util.get_cls_kwargs(dialect_cls).union(
            getattr(dialect_cls, "init_kwargs", set())
        ):
            if k in kwargs:
                dialect_args[k] = pop_kwarg(k)

        kwargs.pop("module", None)  # unused
        dbapi_args = {}
        for k in util.get_func_kwargs(dialect_cls.dbapi):
            if k in kwargs:
                dbapi_args[k] = pop_kwarg(k)
        dbapi = dialect_cls.dbapi(**dbapi_args)
        dialect_args["dbapi"] = dbapi

        dialect = dialect_cls(**dialect_args)
        pool_class = kwargs.pop("pool_class", None)
        pool = await dialect.init_pool(u, loop, pool_class=pool_class)

        engine_args = dict(loop=loop)
        for k in util.get_cls_kwargs(engine_cls):
            if k in kwargs:
                engine_args[k] = pop_kwarg(k)

        # all kwargs should be consumed
        if kwargs:
            raise TypeError(
                "Invalid argument(s) %s sent to create_engine(), "
                "using configuration %s/%s.  Please check that the "
                "keyword arguments are appropriate for this combination "
                "of components."
                % (
                    ",".join("'%s'" % k for k in kwargs),
                    dialect_cls.__name__,
                    engine_cls.__name__,
                )
            )

        engine = engine_cls(dialect, pool, **engine_args)

        dialect_cls.engine_created(engine)

        return engine


GinoStrategy()



class _Break(BaseException):
    def __init__(self, tx, commit):
        super().__init__()
        self.tx = tx
        self.commit = commit


class GinoTransaction:
    """
    Represents an underlying database transaction and its connection, offering
    methods to manage this transaction.

    :class:`.GinoTransaction` is supposed to be created by either
    :meth:`gino.engine.GinoConnection.transaction`, or
    :meth:`gino.engine.GinoEngine.transaction`, or
    :meth:`gino.api.Gino.transaction`, shown as follows::

        async with db.transaction() as tx:
            ...

        async with engine.transaction() as tx:
            ...

        async with conn.transaction() as tx:
            ...

        tx = await conn.transaction()
        try:
            ...
            await tx.commit()
        except Exception:
            await tx.rollback()
            raise

    When in use with asynchronous context manager, :class:`.GinoTransaction`
    will be in **managed** mode, while the last example with ``await`` will put
    the :class:`.GinoTransaction` in **manual** mode where you have to call
    the :meth:`.commit` or :meth:`.rollback` to manually close the transaction.

    In **managed** mode the transaction will be automatically committed or
    rolled back on exiting the ``async with`` block depending on whether there
    is an exception or not. Meanwhile, you can explicitly exit the transaction
    early by :meth:`.raise_commit` or :meth:`.raise_rollback` which will raise
    an internal exception managed by the asynchronous context manager and
    interpreted as a commit or rollback action. In a nested transaction
    situation, the two exit-early methods always close up the very transaction
    which the two methods are referenced upon - all children transactions are
    either committed or rolled back correspondingly, while no parent
    transaction was ever touched. For example::

        async with db.transaction() as tx1:
            async with db.transaction() as tx2:
                async with db.transaction() as tx3:
                    tx2.raise_rollback()
                    # Won't reach here
                # Won't reach here
            # Continues here with tx1, with both tx2 and tx3 rolled back.

            # For PostgreSQL, tx1 can still be committed successfully because
            # tx2 and tx3 are just SAVEPOINTs in transaction tx1

    .. tip::

        The internal exception raised from :meth:`.raise_commit` and
        :meth:`.raise_rollback` is a subclass of :exc:`BaseException`, so
        normal ``try ... except Exception:`` can't trap the commit or rollback.

    """

    def __init__(self, conn, args, kwargs):
        self._conn = conn
        self._args = args
        self._kwargs = kwargs
        self._tx = None
        self._managed = None

    async def _begin(self):
        raw_conn = await self._conn.get_raw_connection()
        self._tx = self._conn.dialect.transaction(raw_conn, self._args, self._kwargs)
        await self._tx.begin()
        return self

    @property
    def connection(self):
        """
        Accesses to the :class:`~gino.engine.GinoConnection` of this
        transaction. This is useful if when the transaction is started from
        ``db`` or ``engine`` where the connection is implicitly acquired for
        you together with the transaction.

        """
        return self._conn

    @property
    def raw_transaction(self):
        """
        Accesses to the underlying transaction object, whose type depends on
        the dialect in use.

        """
        return self._tx.raw_transaction

    def raise_commit(self):
        """
        Only available in managed mode: skip rest of the code in this
        transaction and commit immediately by raising an internal exception,
        which will be caught and handled by the asynchronous context manager::

            async with db.transaction() as tx:
                await user.update(age=64).apply()
                tx.raise_commit()
                await user.update(age=32).apply()  # won't reach here

            assert user.age == 64  # no exception raised before

        """
        if not self._managed:
            raise AssertionError("Illegal in manual mode, use `commit` instead.")
        raise _Break(self, True)

    async def commit(self):
        """
        Only available in manual mode: manually commit this transaction.

        """
        if self._managed:
            raise AssertionError(
                "Illegal in managed mode, " "use `raise_commit` instead."
            )
        await self._tx.commit()

    def raise_rollback(self):
        """
        Only available in managed mode: skip rest of the code in this
        transaction and rollback immediately by raising an internal exception,
        which will be caught and handled by the asynchronous context manager::

            assert user.age == 64  # assumption

            async with db.transaction() as tx:
                await user.update(age=32).apply()
                tx.raise_rollback()
                await user.update(age=128).apply()  # won't reach here

            assert user.age == 64  # no exception raised before

        """
        if not self._managed:
            raise AssertionError("Illegal in manual mode, use `rollback` instead.")
        raise _Break(self, False)

    async def rollback(self):
        """
        Only available in manual mode: manually rollback this transaction.

        """
        if self._managed:
            raise AssertionError(
                "Illegal in managed mode, " "use `raise_rollback` instead."
            )
        await self._tx.rollback()

    def __await__(self):
        if self._managed is not None:
            raise AssertionError("Cannot start the same transaction twice")
        self._managed = False
        return self._begin().__await__()

    async def __aenter__(self):
        if self._managed is not None:
            raise AssertionError("Cannot start the same transaction twice")
        self._managed = True
        await self._begin()
        return self

    async def __aexit__(self, ex_type, ex, ex_tb):
        is_break = ex_type is _Break
        if is_break and ex.commit or ex_type is None:
            await self._tx.commit()
        else:
            await self._tx.rollback()
        if is_break and ex.tx is self:
            return True



# noinspection PyProtectedMember
from sqlalchemy import exc, util
from sqlalchemy.sql.base import _bind_or_error
from sqlalchemy.sql.ddl import (
    AddConstraint,
    CreateIndex,
    CreateSequence,
    CreateTable,
    DropConstraint,
    DropIndex,
    DropSequence,
    DropTable,
    SchemaDropper,
    SchemaGenerator,
    SetColumnComment,
    SetTableComment,
    sort_tables_and_constraints,
)
from sqlalchemy.types import SchemaType


class AsyncVisitor:
    async def traverse_single(self, obj, **kw):
        # Support both SQLAlchemy 1.2 and 1.3
        for v in getattr(
            self, "visitor_iterator", getattr(self, "_visitor_iterator", None)
        ):
            meth = getattr(v, "visit_%s" % obj.__visit_name__, None)
            if meth:
                return await meth(obj, **kw)


class AsyncSchemaGenerator(AsyncVisitor, SchemaGenerator):
    async def _can_create_table(self, table):
        self.dialect.validate_identifier(table.name)
        effective_schema = self.connection.schema_for_object(table)
        if effective_schema:
            self.dialect.validate_identifier(effective_schema)
        return not self.checkfirst or not (
            await self.dialect.has_table(
                self.connection, table.name, schema=effective_schema
            )
        )

    async def _can_create_sequence(self, sequence):
        effective_schema = self.connection.schema_for_object(sequence)

        return self.dialect.supports_sequences and (
            (not self.dialect.sequences_optional or not sequence.optional)
            and (
                not self.checkfirst
                or not await self.dialect.has_sequence(
                    self.connection, sequence.name, schema=effective_schema
                )
            )
        )

    async def visit_metadata(self, metadata):
        if self.tables is not None:
            tables = self.tables
        else:
            tables = list(metadata.tables.values())

        tables_create = []
        for t in tables:
            if await self._can_create_table(t):
                tables_create.append(t)

        collection = sort_tables_and_constraints(tables_create)

        seq_coll = []
        # noinspection PyProtectedMember
        for s in metadata._sequences.values():
            if s.column is None and await self._can_create_sequence(s):
                seq_coll.append(s)

        event_collection = [t for (t, fks) in collection if t is not None]
        await _Async(metadata.dispatch.before_create)(
            metadata,
            self.connection,
            tables=event_collection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
        )

        for seq in seq_coll:
            await self.traverse_single(seq, create_ok=True)

        for table, fkcs in collection:
            if table is not None:
                await self.traverse_single(
                    table,
                    create_ok=True,
                    include_foreign_key_constraints=fkcs,
                    _is_metadata_operation=True,
                )
            else:
                for fkc in fkcs:
                    await self.traverse_single(fkc)

        await _Async(metadata.dispatch.after_create)(
            metadata,
            self.connection,
            tables=event_collection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
        )

    async def visit_table(
        self,
        table,
        create_ok=False,
        include_foreign_key_constraints=None,
        _is_metadata_operation=False,
    ):
        if not create_ok and not await self._can_create_table(table):
            return

        await _Async(table.dispatch.before_create)(
            table,
            self.connection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
            _is_metadata_operation=_is_metadata_operation,
        )

        for column in table.columns:
            if column.default is not None:
                await self.traverse_single(column.default)

        if not self.dialect.supports_alter:
            # e.g., don't omit any foreign key constraints
            include_foreign_key_constraints = None

        await self.connection.status(
            CreateTable(
                table, include_foreign_key_constraints=include_foreign_key_constraints
            )
        )

        if hasattr(table, "indexes"):
            for index in table.indexes:
                await self.traverse_single(index)

        if self.dialect.supports_comments and not self.dialect.inline_comments:
            if table.comment is not None:
                await self.connection.status(SetTableComment(table))

            for column in table.columns:
                if column.comment is not None:
                    await self.connection.status(SetColumnComment(column))

        await _Async(table.dispatch.after_create)(
            table,
            self.connection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
            _is_metadata_operation=_is_metadata_operation,
        )

    async def visit_foreign_key_constraint(self, constraint):
        if not self.dialect.supports_alter:
            return
        await self.connection.status(AddConstraint(constraint))

    async def visit_sequence(self, sequence, create_ok=False):
        if not create_ok and not await self._can_create_sequence(sequence):
            return
        await self.connection.status(CreateSequence(sequence))

    async def visit_index(self, index):
        await self.connection.status(CreateIndex(index))


class AsyncSchemaDropper(AsyncVisitor, SchemaDropper):
    async def visit_metadata(self, metadata):
        if self.tables is not None:
            tables = self.tables
        else:
            tables = list(metadata.tables.values())

        try:
            unsorted_tables = []
            for t in tables:
                if await self._can_drop_table(t):
                    unsorted_tables.append(t)
            collection = list(
                reversed(
                    sort_tables_and_constraints(
                        unsorted_tables,
                        filter_fn=lambda constraint: False
                        if not self.dialect.supports_alter or constraint.name is None
                        else None,
                    )
                )
            )
        except exc.CircularDependencyError as err2:
            if not self.dialect.supports_alter:
                util.warn(
                    "Can't sort tables for DROP; an "
                    "unresolvable foreign key "
                    "dependency exists between tables: %s, and backend does "
                    "not support ALTER.  To restore at least a partial sort, "
                    "apply use_alter=True to ForeignKey and "
                    "ForeignKeyConstraint "
                    "objects involved in the cycle to mark these as known "
                    "cycles that will be ignored."
                    % (", ".join(sorted([t.fullname for t in err2.cycles])))
                )
                collection = [(t, ()) for t in unsorted_tables]
            else:
                util.raise_from_cause(
                    exc.CircularDependencyError(
                        err2.args[0],
                        err2.cycles,
                        err2.edges,
                        msg="Can't sort tables for DROP; an unresolvable "
                        "foreign key dependency exists between tables: %s."
                        "  Please ensure that the ForeignKey and "
                        "ForeignKeyConstraint objects involved in the "
                        "cycle have names so that they can be dropped "
                        "using DROP CONSTRAINT."
                        % (", ".join(sorted([t.fullname for t in err2.cycles]))),
                    )
                )

        seq_coll = []
        for s in metadata._sequences.values():
            if s.column is None and await self._can_drop_sequence(s):
                seq_coll.append(s)

        event_collection = [t for (t, fks) in collection if t is not None]

        await _Async(metadata.dispatch.before_drop)(
            metadata,
            self.connection,
            tables=event_collection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
        )

        for table, fkcs in collection:
            if table is not None:
                await self.traverse_single(
                    table, drop_ok=True, _is_metadata_operation=True
                )
            else:
                for fkc in fkcs:
                    await self.traverse_single(fkc)

        for seq in seq_coll:
            await self.traverse_single(seq, drop_ok=True)

        await _Async(metadata.dispatch.after_drop)(
            metadata,
            self.connection,
            tables=event_collection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
        )

    async def _can_drop_table(self, table):
        self.dialect.validate_identifier(table.name)
        effective_schema = self.connection.schema_for_object(table)
        if effective_schema:
            self.dialect.validate_identifier(effective_schema)
        return not self.checkfirst or (
            await self.dialect.has_table(
                self.connection, table.name, schema=effective_schema
            )
        )

    async def _can_drop_sequence(self, sequence):
        effective_schema = self.connection.schema_for_object(sequence)
        return self.dialect.supports_sequences and (
            (not self.dialect.sequences_optional or not sequence.optional)
            and (
                not self.checkfirst
                or await self.dialect.has_sequence(
                    self.connection, sequence.name, schema=effective_schema
                )
            )
        )

    async def visit_index(self, index):
        await self.connection.status(DropIndex(index))

    async def visit_table(self, table, drop_ok=False, _is_metadata_operation=False):
        if not drop_ok and not await self._can_drop_table(table):
            return

        await _Async(table.dispatch.before_drop)(
            table,
            self.connection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
            _is_metadata_operation=_is_metadata_operation,
        )

        for column in table.columns:
            if column.default is not None:
                await self.traverse_single(column.default)

        await self.connection.status(DropTable(table))

        await _Async(table.dispatch.after_drop)(
            table,
            self.connection,
            checkfirst=self.checkfirst,
            _ddl_runner=self,
            _is_metadata_operation=_is_metadata_operation,
        )

    async def visit_foreign_key_constraint(self, constraint):
        if not self.dialect.supports_alter:
            return
        await self.connection.status(DropConstraint(constraint))

    async def visit_sequence(self, sequence, drop_ok=False):
        if not drop_ok and not await self._can_drop_sequence(sequence):
            return
        await self.connection.status(DropSequence(sequence))


class GinoSchemaVisitor:
    __slots__ = ("_item",)

    def __init__(self, item):
        self._item = item

    async def create(self, bind=None, *args, **kwargs):
        if bind is None:
            bind = _bind_or_error(self._item)
        await getattr(bind, "_run_visitor")(
            AsyncSchemaGenerator, self._item, *args, **kwargs
        )
        return self._item

    async def drop(self, bind=None, *args, **kwargs):
        if bind is None:
            bind = _bind_or_error(self._item)
        await getattr(bind, "_run_visitor")(
            AsyncSchemaDropper, self._item, *args, **kwargs
        )

    async def create_all(self, bind=None, tables=None, checkfirst=True):
        await self.create(bind=bind, tables=tables, checkfirst=checkfirst)

    async def drop_all(self, bind=None, tables=None, checkfirst=True):
        await self.drop(bind=bind, tables=tables, checkfirst=checkfirst)


class AsyncSchemaTypeMixin:
    async def create_async(self, bind=None, checkfirst=False):
        if bind is None:
            bind = _bind_or_error(self)
        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await t.create_async(bind=bind, checkfirst=checkfirst)

    async def drop_async(self, bind=None, checkfirst=False):
        if bind is None:
            bind = _bind_or_error(self)
        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await t.drop_async(bind=bind, checkfirst=checkfirst)

    async def _on_table_create_async(self, target, bind, **kw):
        if not self._is_impl_for_variant(bind.dialect, kw):
            return

        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await getattr(t, "_on_table_create_async")(target, bind, **kw)

    async def _on_table_drop_async(self, target, bind, **kw):
        if not self._is_impl_for_variant(bind.dialect, kw):
            return

        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await getattr(t, "_on_table_drop_async")(target, bind, **kw)

    async def _on_metadata_create_async(self, target, bind, **kw):
        if not self._is_impl_for_variant(bind.dialect, kw):
            return

        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await getattr(t, "_on_metadata_create_async")(target, bind, **kw)

    async def _on_metadata_drop_async(self, target, bind, **kw):
        if not self._is_impl_for_variant(bind.dialect, kw):
            return

        t = self.dialect_impl(bind.dialect)
        if t.__class__ is not self.__class__ and isinstance(t, SchemaType):
            await getattr(t, "_on_metadata_drop_async")(target, bind, **kw)


async def _call_portable_instancemethod(fn, args, kw):
    m = getattr(fn.target, fn.name + "_async", None)
    if m is None:
        return fn(*args, **kw)
    else:
        kw.update(fn.kwargs)
        return await m(*args, **kw)


class _Async:
    def __init__(self, listener):
        self._listener = listener

    async def call(self, *args, **kw):
        for fn in self._listener.parent_listeners:
            await _call_portable_instancemethod(fn, args, kw)
        for fn in self._listener.listeners:
            await _call_portable_instancemethod(fn, args, kw)

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)


def patch_schema(db):
    for st in {"Enum"}:
        setattr(db, st, type(st, (getattr(db, st), AsyncSchemaTypeMixin), {}))



import asyncio
import contextvars
import sys
import types


def patch_asyncio():
    """Patches asyncio to support :mod:`contextvars`.

    This is automatically called when :mod:`gino` is imported. If Python version is 3.7
    or greater, this function is a no-op.
    """

    if not sys.version_info < (3, 7):
        return

    def _get_context():
        state = _get_state()
        ctx = getattr(state, "context", None)
        if ctx is None:
            ctx = contextvars.Context()
            state.context = ctx
        return ctx

    def _set_context(ctx):
        state = _get_state()
        state.context = ctx

    def _get_state():
        loop = asyncio._get_running_loop()
        if loop is None:
            return contextvars._state
        task = asyncio.Task.current_task(loop=loop)
        return contextvars._state if task is None else task

    contextvars._get_context = _get_context
    contextvars._set_context = _set_context

    def create_task(loop, coro):
        task = loop._orig_create_task(coro)
        if task._source_traceback:
            del task._source_traceback[-1]
        task.context = contextvars.copy_context()
        return task

    def _patch_loop(loop):
        if loop and not hasattr(loop, "_orig_create_task"):
            loop._orig_create_task = loop.create_task
            loop.create_task = types.MethodType(create_task, loop)
        return loop

    def get_event_loop():
        return _patch_loop(_get_event_loop())

    def set_event_loop(loop):
        return _set_event_loop(_patch_loop(loop))

    def new_event_loop():
        return _patch_loop(_new_event_loop())

    _get_event_loop = asyncio.get_event_loop
    _set_event_loop = asyncio.set_event_loop
    _new_event_loop = asyncio.new_event_loop

    asyncio.get_event_loop = asyncio.events.get_event_loop = get_event_loop
    asyncio.set_event_loop = asyncio.events.set_event_loop = set_event_loop
    asyncio.new_event_loop = asyncio.events.new_event_loop = new_event_loop



from datetime import datetime

import sqlalchemy as sa

from .exceptions import UnknownJSONPropertyError

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
NONE = object()


class Hook:
    def __init__(self, parent):
        self.parent = parent
        self.method = None

    def __call__(self, method):
        self.method = method
        return self.parent

    def call(self, instance, val):
        if self.method is not None:
            val = self.method(instance, val)
        return val


class JSONProperty:
    def __init__(self, default=None, prop_name="profile"):
        self.name = None
        self.default = default
        self.prop_name = prop_name
        self.expression = Hook(self)
        self.after_get = Hook(self)
        self.before_set = Hook(self)

    def __get__(self, instance, owner):
        if instance is None:
            exp = self.make_expression(getattr(owner, self.prop_name)[self.name])
            return self.expression.call(owner, exp)
        val = self.get_profile(instance).get(self.name, NONE)
        if val is NONE:
            if callable(self.default):
                val = self.default(instance)
            else:
                val = self.default
        return self.after_get.call(instance, val)

    def __set__(self, instance, value):
        self.get_profile(instance)[self.name] = self.before_set.call(instance, value)

    def __delete__(self, instance):
        self.get_profile(instance).pop(self.name, None)

    def get_profile(self, instance):
        if instance.__profile__ is None:
            props = type(instance).__dict__
            instance.__profile__ = {}
            profiles = {}
            for prop_name in getattr(instance, "__json_prop_names__", set()):
                profiles.update(getattr(instance, prop_name, None) or {})
            for key, value in profiles.items():
                if key not in props:
                    raise UnknownJSONPropertyError(
                        "`{}` is found in `{}` of instance {}, "
                        "but it is not defined".format(key, self.prop_name, instance)
                    )

                prop = props[key]
                if not isinstance(prop, JSONProperty):
                    raise UnknownJSONPropertyError(
                        "`{}` is found in `{}` of instance {}, "
                        "but column `{}` is not an instance of the "
                        "`JSONProperty` type.".format(
                            key, self.prop_name, instance, key
                        )
                    )
                instance.__profile__[key] = prop.decode(value)

        return instance.__profile__

    def save(self, instance, value=NONE):
        profile = getattr(instance, self.prop_name, None)
        if profile is None:
            profile = {}
            setattr(instance, self.prop_name, profile)
        if value is NONE:
            value = instance.__profile__[self.name]
        if not isinstance(value, sa.sql.ClauseElement):
            value = self.encode(value)
        rv = profile[self.name] = value
        return rv

    def reload(self, instance):
        if instance.__profile__ is None:
            return
        profile = getattr(instance, self.prop_name, None) or {}
        value = profile.get(self.name, NONE)
        if value is NONE:
            instance.__profile__.pop(self.name, None)
        else:
            instance.__profile__[self.name] = self.decode(value)

    def make_expression(self, base_exp):
        return base_exp

    def decode(self, val):
        return val

    def encode(self, val):
        return val

    def __hash__(self):
        return hash(self.name)


class StringProperty(JSONProperty):
    def make_expression(self, base_exp):
        return base_exp.astext


class DateTimeProperty(JSONProperty):
    def make_expression(self, base_exp):
        return base_exp.astext.cast(sa.DateTime)

    def decode(self, val):
        if val:
            val = datetime.strptime(val, DATETIME_FORMAT)
        return val

    def encode(self, val):
        if isinstance(val, datetime):
            val = val.strftime(DATETIME_FORMAT)
        return val


class IntegerProperty(JSONProperty):
    def make_expression(self, base_exp):
        return base_exp.astext.cast(sa.Integer)

    def decode(self, val):
        if val is not None:
            val = int(val)
        return val

    def encode(self, val):
        if val is not None:
            val = int(val)
        return val


class BooleanProperty(JSONProperty):
    def make_expression(self, base_exp):
        return base_exp.astext.cast(sa.Boolean)

    def decode(self, val):
        if val is not None:
            val = bool(val)
        return val

    def encode(self, val):
        if val is not None:
            val = bool(val)
        return val


class ObjectProperty(JSONProperty):
    def decode(self, val):
        if val is not None:
            val = dict(val)
        return val

    def encode(self, val):
        if val is not None:
            val = dict(val)
        return val


class ArrayProperty(JSONProperty):
    def decode(self, val):
        if val is not None:
            val = list(val)
        return val

    def encode(self, val):
        if val is not None:
            val = list(val)
        return val


__all__ = [
    "JSONProperty",
    "StringProperty",
    "DateTimeProperty",
    "IntegerProperty",
    "BooleanProperty",
    "ObjectProperty",
    "ArrayProperty",
]



"""
Namespace package for GINO extensions.

This namespace package didn't use any of the `3 official solutions
<https://packaging.python.org/guides/packaging-namespace-packages/>`__. Instead,
``gino.ext`` adds a hook into :data:`sys.meta_path`, and utilize the :ref:`entry-points`
to find the extensions.

Any GINO extension package should provide an entry point like this::

    [gino.extensions]
    starlette = gino_starlette

So that the Python package ``gino_starlette`` will also be importable through
``gino.ext.starlette``.
"""

import sys
from importlib.abc import MetaPathFinder, Loader
from importlib.machinery import ModuleSpec
from importlib.util import find_spec

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


class _GinoExtensionCompatProxyLoader(Loader):
    def __init__(self, fullname, loader):
        self._fullname = fullname
        self._loader = loader

    def create_module(self, spec):
        return self._loader.create_module(spec)

    def exec_module(self, mod):
        sys.modules[self._fullname] = mod
        return self._loader.exec_module(mod)


class _GinoExtensionCompatNoopLoader(Loader):
    def __init__(self, mod):
        self._mod = mod

    def create_module(self, spec):
        return self._mod

    def exec_module(self, mod):
        pass


class _GinoExtensionCompatFinder(MetaPathFinder):
    def __init__(self):
        self._redirects = {
            __name__ + "." + ep.name: ep.value
            for ep in entry_points().get("gino.extensions", [])
        }

    # noinspection PyUnusedLocal
    def find_spec(self, fullname, path, target=None):
        target = self._redirects.get(fullname)
        if target:
            mod = sys.modules.get(target)
            if mod is None:
                spec = find_spec(target)
                spec.loader = _GinoExtensionCompatProxyLoader(fullname, spec.loader)
                return spec
            else:
                return ModuleSpec(fullname, _GinoExtensionCompatNoopLoader(mod))
        elif fullname.startswith(__name__):
            raise ImportError(
                "Cannot import {} - is gino-{} a valid extension and installed?".format(
                    fullname, fullname[len(__name__) + 1 :]
                )
            )

    @classmethod
    def uninstall(cls):
        if sys.meta_path:
            for i in range(len(sys.meta_path) - 1, -1, -1):
                if type(sys.meta_path[i]).__name__ == cls.__name__:
                    del sys.meta_path[i]

    def install(self):
        self.uninstall()
        sys.meta_path.append(self)


_GinoExtensionCompatFinder().install()



import inspect
import itertools
import weakref

import sqlalchemy as sa
from sqlalchemy.sql import ClauseElement

from . import json_support
from .declarative import Model, InvertDict
from .exceptions import NoSuchRowError
from .loader import AliasLoader, ModelLoader

DEFAULT = object()


class _Create:
    def __get__(self, instance, owner):
        if instance is None:
            # noinspection PyProtectedMember
            return owner._create_without_instance
        else:
            # noinspection PyProtectedMember
            return instance._create


class _Query:
    def __get__(self, instance, owner):
        # noinspection PyProtectedMember
        owner._check_abstract()
        q = sa.select([owner.__table__])
        if instance is not None:
            q = q.where(instance.lookup())
        return q.execution_options(model=weakref.ref(owner))


class _Select:
    def __get__(self, instance, owner):
        def select(*args):
            q = sa.select([getattr(owner, x) for x in args])
            if instance is not None:
                q = q.where(instance.lookup())
            return q.execution_options(model=weakref.ref(owner), return_model=False)

        return select


class _Update:
    def __get__(self, instance, owner):
        if instance is None:
            # noinspection PyProtectedMember
            owner._check_abstract()
            q = owner.__table__.update()
            return q.execution_options(model=weakref.ref(owner))
        else:
            # noinspection PyProtectedMember
            return instance._update


class _Delete:
    def __get__(self, instance, owner):
        if instance is None:
            # noinspection PyProtectedMember
            owner._check_abstract()
            q = owner.__table__.delete()
            return q.execution_options(model=weakref.ref(owner))
        else:
            # noinspection PyProtectedMember
            return instance._delete


class UpdateRequest:
    """
    A collection of attributes and their new values to update on one model
    instance.

    :class:`.UpdateRequest` instances are created by :attr:`.CRUDModel.update`,
    don't instantiate manually unless required. Every :class:`.UpdateRequest`
    instance is bound to one model instance, all updates are for that one
    specific model instance and its database row.

    """

    def __init__(self, instance: "CRUDModel"):
        self._instance = instance
        self._values = {}
        self._props = {}
        self._literal = True
        self._locator = None
        if instance.__table__ is not None:
            try:
                self._locator = instance.lookup()
            except LookupError:
                # apply() will fail anyway, but still allow update()
                pass

    def _set(self, key, value):
        self._values[key] = value

    def _set_prop(self, prop, value):
        if isinstance(value, ClauseElement):
            self._literal = False
        self._props[prop] = value

    async def apply(self, bind=None, timeout=DEFAULT):
        """
        Apply pending updates into database by executing an ``UPDATE`` SQL.

        :param bind: A :class:`~gino.engine.GinoEngine` to execute the SQL, or
          ``None`` (default) to use the bound engine in the metadata.

        :param timeout: Seconds to wait for the database to finish executing,
          ``None`` for wait forever. By default it will use the ``timeout``
          execution option value if unspecified.

        :return: ``self`` for chaining calls.

        """
        if self._locator is None:
            raise TypeError(
                "Model {} has no table, primary key or custom lookup()".format(
                    self._instance.__class__.__name__
                )
            )
        cls = type(self._instance)
        values = self._values.copy()

        # handle JSON columns
        json_updates = {}
        for prop, value in self._props.items():
            value = prop.save(self._instance, value)
            updates = json_updates.setdefault(prop.prop_name, {})
            if self._literal:
                updates[prop.name] = value
            else:
                if isinstance(value, int):
                    value = sa.cast(value, sa.BigInteger)
                elif not isinstance(value, ClauseElement):
                    value = sa.cast(value, sa.Unicode)
                updates[sa.cast(prop.name, sa.Unicode)] = value
        for prop_name, updates in json_updates.items():
            prop = getattr(cls, prop_name)
            from .dialects.asyncpg import JSONB

            if isinstance(prop.type, JSONB):
                if self._literal:
                    values[prop_name] = prop.concat(updates)
                else:
                    values[prop_name] = prop.concat(
                        sa.func.jsonb_build_object(*itertools.chain(*updates.items()))
                    )
            else:
                raise TypeError(
                    "{} is not supported to update json "
                    "properties in Gino. Please consider using "
                    "JSONB.".format(prop.type)
                )

        opts = dict(return_model=False)
        if timeout is not DEFAULT:
            opts["timeout"] = timeout
        clause = (
            type(self._instance)
            .update.where(self._locator,)
            .values(**self._instance._get_sa_values(values),)
            .returning(*[getattr(cls, key) for key in values],)
            .execution_options(**opts)
        )
        if bind is None:
            bind = cls.__metadata__.bind
        row = await bind.first(clause)
        if not row:
            raise NoSuchRowError()
        for k, v in row.items():
            self._instance.__values__[self._instance._column_name_map.invert_get(k)] = v
        for prop in self._props:
            prop.reload(self._instance)
        return self

    def update(self, **values):
        """
        Set given attributes on the bound model instance, and add them into
        the update collections for :meth:`.apply`.

        Given keyword-only arguments are pairs of attribute names and values to
        update. This is not a coroutine, calling :meth:`.update` will have
        instant effect on the bound model instance - its in-memory values will
        be updated immediately. Therefore this can be used individually as a
        shortcut to update several attributes in a batch::

            user.update(age=32, disabled=True)

        :meth:`.update` returns ``self`` for chaining calls to either
        :meth:`.apply` or another :meth:`.update`. If one attribute is updated
        several times by the same :class:`.UpdateRequest`, then only the last
        value is remembered for :meth:`.apply`.

        Updated values can be SQLAlchemy expressions, for example an atomic
        increment for user balance looks like this::

            await user.update(balance=User.balance + 100).apply()

        .. note::

            Expression values will not affect the in-memory attribute value on
            :meth:`.update` before :meth:`.apply`, because it has no knowledge
            of the latest value in the database. After :meth:`.apply` the new
            value will be automatically reloaded from database with
            ``RETURNING`` clause.

        """

        cls = type(self._instance)
        for key, value in values.items():
            prop = cls.__dict__.get(key)
            if isinstance(prop, json_support.JSONProperty):
                value_from = "__profile__"
                method = self._set_prop
                k = prop
            else:
                value_from = "__values__"
                method = self._set
                k = key
            if not isinstance(value, ClauseElement):
                setattr(self._instance, key, value)
                value = getattr(self._instance, value_from)[key]
            method(k, value)
        return self


class Alias:
    """
    Experimental proxy for table alias on model.

    """

    def __init__(self, model, *args, **kwargs):
        # noinspection PyProtectedMember
        model._check_abstract()
        self.model = model
        self.alias = model.__table__.alias(*args, **kwargs)

    def __getattr__(self, item):
        rv = getattr(
            self.alias.columns,
            item,
            getattr(self.alias, item, getattr(self.model, item, DEFAULT)),
        )
        if rv is DEFAULT:
            raise AttributeError
        return rv

    def __iter__(self):
        return iter(self.alias.columns)

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def load(self, *column_names, **relationships):
        return AliasLoader(self, *column_names, **relationships)

    def on(self, on_clause):
        return self.load().on(on_clause)

    def distinct(self, *columns):
        return self.load().distinct(*columns)


# noinspection PyProtectedMember
@sa.inspection._inspects(Alias)
def _inspect_alias(target):
    return sa.inspection.inspect(target.alias)


class CRUDModel(Model):
    """
    The base class for models with CRUD support.

    Don't inherit from this class directly, because it has no metadata. Use
    :attr:`db.Model <gino.api.Gino.Model>` instead.

    """

    create = _Create()
    """
    This ``create`` behaves a bit different on model classes compared to model
    instances.

    On model classes, ``create`` will create a new model instance and insert
    it into database. On model instances, ``create`` will just insert the
    instance into the database.

    Under the hood :meth:`.create` uses ``INSERT ... RETURNING ...`` to
    create the new model instance and load it with database default data if
    not specified.

    Some examples::

        user1 = await User.create(name='fantix', age=32)
        user2 = User(name='gino', age=42)
        await user2.create()

    :param bind: A :class:`~gino.engine.GinoEngine` to execute the
      ``INSERT`` statement with, or ``None`` (default) to use the bound
      engine on the metadata (:class:`~gino.api.Gino`).

    :param timeout: Seconds to wait for the database to finish executing,
      ``None`` for wait forever. By default it will use the ``timeout``
      execution option value if unspecified.

    :param values: Keyword arguments are pairs of attribute names and their
      initial values. Only available when called on a model class.

    :return: The instance of this model class (newly created or existing).

    """

    query = _Query()
    """
    Get a SQLAlchemy query clause of the table behind this model. This equals
    to :func:`sqlalchemy.select([self.__table__])
    <sqlalchemy.sql.expression.select>`. If this attribute is retrieved on a
    model instance, then a where clause to locate this instance by its primary
    key is appended to the returning query clause. This model type is set as
    the execution option ``model`` in the returning clause, so by default the
    query yields instances of this model instead of database rows.

    """

    update = _Update()
    """
    This ``update`` behaves quite different on model classes rather than model
    instances.

    On model classes, ``update`` is an attribute of type
    :class:`~sqlalchemy.sql.expression.Update` for massive updates, for
    example::

        await User.update.values(enabled=True).where(...).gino.status()

    Like :attr:`.query`, the update query also has the ``model`` execution
    option of this model, so if you use the
    :meth:`~sqlalchemy.sql.expression.Update.returning` clause, the query shall
    return model objects.

    However on model instances, ``update()`` is a method which accepts keyword
    arguments only and returns an :class:`.UpdateRequest` to update this single
    model instance. The keyword arguments are pairs of attribute names and new
    values. This is the same as :meth:`.UpdateRequest.update`, feel free to
    read more about it. A normal usage example would be like this::

        await user.update(name='new name', age=32).apply()

    Here, the :meth:`await ... apply() <.UpdateRequest.apply>` executes the
    actual ``UPDATE`` SQL in the database, while ``user.update()`` only makes
    changes in the memory, and collect all changes into an
    :class:`.UpdateRequest` instance.

    """

    delete = _Delete()
    """
    Similar to :meth:`.update`, this ``delete`` is also different on model
    classes than on model instances.

    On model classes ``delete`` is an attribute of type
    :class:`~sqlalchemy.sql.expression.Delete` for massive deletes, for
    example::

        await User.delete.where(User.enabled.is_(False)).gino.status()

    Similarly you can add a :meth:`~sqlalchemy.sql.expression.Delete.returning`
    clause to the query and it shall return the deleted rows as model objects.

    And on model instances, ``delete()`` is a method to remove the
    corresponding row in the database of this model instance. and returns the
    status returned from the database::

        print(await user.delete())  # e.g. prints DELETE 1

    .. note::

        ``delete()`` only removes the row from database, it does not affect the
        current model instance.

    :param bind: An optional :class:`~gino.engine.GinoEngine` if current
      metadata (:class:`~gino.api.Gino`) has no bound engine, or specifying a
      different :class:`~gino.engine.GinoEngine` to execute the ``DELETE``.

    :param timeout: Seconds to wait for the database to finish executing,
      ``None`` for wait forever. By default it will use the ``timeout``
      execution option value if unspecified.

    """

    select = _Select()
    """
    Build a query to retrieve only specified columns from this table.

    This method accepts positional string arguments as names of attributes to
    retrieve, and returns a :class:`~sqlalchemy.sql.expression.Select` for
    query. The returning query object is always set with two execution options:

    1. ``model`` is set to this model type
    2. ``return_model`` is set to ``False``

    So that by default it always return rows instead of model instances, while
    column types can be inferred correctly by the ``model`` option.

    For example::

        async for row in User.select('id', 'name').gino.iterate():
            print(row['id'], row['name'])

    If :meth:`.select` is invoked on a model instance, then a ``WHERE`` clause
    to locate this instance by its primary key is appended to the returning
    query clause. This is useful when you want to retrieve a latest value of a
    field on current model instance from database::

        db_age = await user.select('age').gino.scalar()

    .. seealso::

        :meth:`~gino.engine.GinoConnection.execution_options`

    """

    _update_request_cls = UpdateRequest
    _column_name_map = InvertDict()

    def __init__(self, **values):
        super().__init__()
        self.__profile__ = None
        self._update_request_cls(self).update(**values)

    @classmethod
    def _init_table(cls, sub_cls):
        rv = Model._init_table(sub_cls)
        if rv is not None:
            rv.__model__ = weakref.ref(sub_cls)
        return rv

    @classmethod
    async def _create_without_instance(cls, bind=None, timeout=DEFAULT, **values):
        return await cls(**values)._create(bind=bind, timeout=timeout)

    async def _create(self, bind=None, timeout=DEFAULT):
        # handle JSON properties
        cls = type(self)
        # noinspection PyUnresolvedReferences,PyProtectedMember
        cls._check_abstract()
        profile_keys = set(self.__profile__.keys() if self.__profile__ else [])
        for key in profile_keys:
            cls.__dict__.get(key).save(self)
        # initialize default values
        for key, prop in cls.__dict__.items():
            if key in profile_keys:
                continue
            if isinstance(prop, json_support.JSONProperty):
                if prop.default is None or prop.after_get.method is not None:
                    continue
                setattr(self, key, getattr(self, key))
                prop.save(self)

        # insert into database
        opts = dict(return_model=False, model=cls)
        if timeout is not DEFAULT:
            opts["timeout"] = timeout
        # noinspection PyArgumentList
        q = (
            cls.__table__.insert()
            .values(**self._get_sa_values(self.__values__))
            .returning(*cls)
            .execution_options(**opts)
        )
        if bind is None:
            bind = cls.__metadata__.bind
        row = await bind.first(q)
        for k, v in row.items():
            self.__values__[self._column_name_map.invert_get(k)] = v
        self.__profile__ = None
        return self

    def _get_sa_values(self, instance_values: dict) -> dict:
        values = {}
        for k, v in instance_values.items():
            values[self._column_name_map[k]] = v
        return values

    @classmethod
    async def get(cls, ident, bind=None, timeout=DEFAULT):
        """
        Get an instance of this model class by primary key.

        For example::

            user = await User.get(request.args.get('user_id'))

        :param ident: Value of the primary key. For composite primary keys this
          should be a tuple of values for all keys in database order, or a dict
          of names (or position numbers in database order starting from zero)
          of all primary keys to their values.

        :param bind: A :class:`~gino.engine.GinoEngine` to execute the
          ``INSERT`` statement with, or ``None`` (default) to use the bound
          engine on the metadata (:class:`~gino.api.Gino`).

        :param timeout: Seconds to wait for the database to finish executing,
          ``None`` for wait forever. By default it will use the ``timeout``
          execution option value if unspecified.

        :return: An instance of this model class, or ``None`` if no such row.

        """
        # noinspection PyUnresolvedReferences,PyProtectedMember
        cls._check_abstract()
        if not isinstance(ident, (list, tuple, dict)):
            ident_ = [ident]
        else:
            ident_ = ident
        columns = cls.__table__.primary_key.columns
        if len(ident_) != len(columns):
            raise ValueError(
                "Incorrect number of values as primary key: "
                "expected {}, got {}.".format(len(columns), len(ident_))
            )
        clause = cls.query
        for i, c in enumerate(columns):
            try:
                val = ident_[i]
            except KeyError:
                val = ident_[cls._column_name_map.invert_get(c.name)]
            clause = clause.where(c == val)
        if timeout is not DEFAULT:
            clause = clause.execution_options(timeout=timeout)
        if bind is None:
            bind = cls.__metadata__.bind
        return await bind.first(clause)

    def append_where_primary_key(self, q):
        """
        Append where clause to locate this model instance by primary on the
        given query, and return the new query.

        This is mostly used internally in GINO, but also available for such
        usage::

            await user.append_where_primary_key(User.query).gino.first()

        which is identical to::

            await user.query.gino.first()

        .. deprecated:: 0.7.6
            Use :meth:`.lookup` instead.

        """
        return q.where(self.lookup())  # pragma: no cover

    def lookup(self):
        """
        Generate where-clause expression to locate this model instance.

        By default this method uses current values of all primary keys, and you
        can override it to behave differently. Most instance-level CRUD
        operations depend on this method internally. Particularly while
        :meth:`.lookup` is called in :meth:`.update`, the where condition is
        used in :meth:`.UpdateRequest.apply`, so that queries like ``UPDATE ...
        SET id = NEW WHERE id = OLD`` could work correctly.

        :return:

        .. versionadded:: 0.7.6

        """
        exps = []
        for c in self.__table__.primary_key.columns:
            exps.append(c == getattr(self, self._column_name_map.invert_get(c.name)))
        if exps:
            return sa.and_(*exps)
        else:
            raise LookupError(
                "Instance-level CRUD operations not allowed on "
                "models without primary keys or lookup(), please"
                " use model-level CRUD operations instead."
            )

    def _update(self, **values):
        return self._update_request_cls(self).update(**values)

    async def _delete(self, bind=None, timeout=DEFAULT):
        cls = type(self)
        # noinspection PyUnresolvedReferences,PyProtectedMember
        cls._check_abstract()
        clause = cls.delete.where(self.lookup())
        if timeout is not DEFAULT:
            clause = clause.execution_options(timeout=timeout)
        if bind is None:
            bind = self.__metadata__.bind
        return (await bind.status(clause))[0]

    def to_dict(self):
        """
        Convenient method to generate a dict from this model instance.

        Keys will be attribute names, while values are loaded from memory (not
        from database). If there are :class:`~gino.json_support.JSONProperty`
        attributes in this model, their source JSON field will not be included
        in the returning dict - instead the JSON attributes will be.

        .. seealso::

            :mod:`.json_support`

        """
        cls = type(self)
        # noinspection PyTypeChecker
        keys = set(cls._column_name_map.invert_get(c.name) for c in cls)
        for key, prop in cls.__dict__.items():
            if isinstance(prop, json_support.JSONProperty):
                keys.add(key)
                keys.discard(prop.prop_name)
        return dict((k, getattr(self, k)) for k in keys)

    @classmethod
    def load(cls, *column_names, **relationships):
        """
        Populates a :class:`.loader.Loader` instance to be used by the
        ``loader`` execution option in order to customize the loading behavior
        to load specified fields into instances of this model.

        The basic usage of this method is to provide the ``loader`` execution
        option (if you are looking for reloading
        the instance from database, check :meth:`.get` or :attr:`.query`) for a
        given query.

        This method takes both positional arguments and keyword arguments with
        very different meanings. The positional arguments should be column
        names as strings, specifying only these columns should be loaded into
        the model instance (other values are discarded even if they are
        retrieved from database). Meanwhile, the keyword arguments should be
        loaders for instance attributes. For example::

            u = await User.query.gino.load(User.load('id', 'name')).first()

        .. tip::

            ``gino.load`` is a shortcut for setting the execution option
            ``loader``.

        This will populate a ``User`` instance with only ``id`` and ``name``
        values, all the rest are simply ``None`` even if the query actually
        returned all the column values.

        ::

            q = User.join(Team).select()
            u = await q.gino.load(User.load(team=Team)).first()

        This will load two instances of model ``User`` and ``Team``, returning
        the ``User`` instance with ``u.team`` set to the ``Team`` instance.

        Both positional and keyword arguments can be used ath the same time. If
        they are both omitted, like ``Team.load()``, it is equivalent to just
        ``Team`` as a loader.

        Additionally, a :class:`.loader.Loader` instance can also be used to
        generate queries, as its structure is usually the same as the query::

            u = await User.load(team=Team).query.gino.first()

        This generates a query like this::

            SELECT users.xxx, ..., teams.xxx, ...
              FROM users LEFT JOIN teams
                ON ...

        The :class:`~.loader.Loader` delegates attributes on the ``query``, so
        ``.query`` can be omitted. The ``LEFT JOIN`` is built-in behavior,
        while the ``ON`` clause is generated based on foreign key. If there is
        no foreign key, or the condition should be customized, you can use
        this::

            u = await User.load(
                team=Team.on(User.team_id == Team.id)).gino.first()

        And you can use both :meth:`~.load` and :meth:`~.on` at the same time
        in a chain, in whatever order suits you.

        .. seealso::

            :meth:`~gino.engine.GinoConnection.execution_options`

        """
        return ModelLoader(cls, *column_names, **relationships)

    @classmethod
    def on(cls, on_clause):
        """
        Customize the on-clause for the auto-generated outer join query.

        .. note::

            This has no effect when provided as the ``loader`` execution option
            for a given query.

        .. seealso::

            :meth:`.load`

        """
        return cls.load().on(on_clause)

    @classmethod
    def distinct(cls, *columns):
        """
        Experimental loader feature to yield only distinct instances by given
        columns.

        """
        return cls.load().distinct(*columns)

    @classmethod
    def none_as_none(cls, enabled=True):
        return cls.load().none_as_none(enabled)

    @classmethod
    def alias(cls, *args, **kwargs):
        """
        Experimental proxy for table alias on model.

        """
        return Alias(cls, *args, **kwargs)

    @classmethod
    def in_query(cls, query):
        """
        Convenient method to get a Model object when using subqueries.

        Though with filters and aggregations, subqueries often return same
        columns as the original table, but SQLAlchemy could not recognize them
        as the columns are in subqueries, so technically they're columns in the
        new "table".

        With this method, the columns are loaded into the original models when
        being used in subqueries. For example::

            query = query.alias('users')
            MyUser = User.in_query(query)

            loader = MyUser.distinct(User1.id).load()
            users = await query.gino.load(loader).all()

        """
        return _get_query_model(cls, query)


def _get_query_model(model, query):
    return QueryModel(model.__name__, (), dict(_model=model, _query=query))


class QueryModel(type):
    """
    Metaclass of Model classes used for subqueries.

    """

    def __getattr__(self, item):
        rv = getattr(
            self._query.columns,
            item,
            getattr(
                self._model.__table__.columns, item, getattr(self._model, item, DEFAULT)
            ),
        )
        # replace `cls` in classmethod in models to `self`
        if inspect.ismethod(rv) and inspect.isclass(rv.__self__):
            return lambda *args, **kwargs: rv.__func__(self, *args, **kwargs)
        if rv is DEFAULT:
            raise AttributeError
        return rv

    def __iter__(self):
        return iter(self._query.columns)

    def __call__(self, *args, **kwargs):
        return self._model(*args, **kwargs)



import inspect
import itertools
import time
import warnings

import asyncpg
from sqlalchemy import util, exc, sql
from sqlalchemy.dialects.postgresql import (  # noqa: F401
    ARRAY,
    CreateEnumType,
    DropEnumType,
    JSON,
    JSONB,
    json,
)
from sqlalchemy.dialects.postgresql.base import (
    ENUM,
    PGCompiler,
    PGDialect,
    PGExecutionContext,
)
from sqlalchemy.sql import sqltypes

from . import base

try:
    import click
except ImportError:
    click = None
JSON_COLTYPE = 114
JSONB_COLTYPE = 3802


class AsyncpgDBAPI(base.BaseDBAPI):
    Error = asyncpg.PostgresError, asyncpg.InterfaceError


class AsyncpgCompiler(PGCompiler):
    @property
    def bindtemplate(self):
        return self._bindtemplate

    @bindtemplate.setter
    def bindtemplate(self, val):
        # noinspection PyAttributeOutsideInit
        self._bindtemplate = val.replace(":", "$")

    def _apply_numbered_params(self):
        if hasattr(self, "string"):
            return super()._apply_numbered_params()


# noinspection PyAbstractClass
class AsyncpgExecutionContext(base.ExecutionContextOverride, PGExecutionContext):
    async def _execute_scalar(self, stmt, type_):
        conn = self.root_connection
        if (
            isinstance(stmt, util.text_type)
            and not self.dialect.supports_unicode_statements
        ):
            stmt = self.dialect._encoder(stmt)[0]

        if self.dialect.positional:
            default_params = self.dialect.execute_sequence_format()
        else:
            default_params = {}

        conn._cursor_execute(self.cursor, stmt, default_params, context=self)
        r = await self.cursor.async_execute(stmt, None, default_params, 1)
        r = r[0][0]
        if type_ is not None:
            # apply type post processors to the result
            proc = type_._cached_result_processor(
                self.dialect, self.cursor.description[0][1]
            )
            if proc:
                return proc(r)
        return r


class AsyncpgIterator:
    def __init__(self, context, iterator):
        self._context = context
        self._iterator = iterator

    async def __anext__(self):
        row = await self._iterator.__anext__()
        return self._context.process_rows([row])[0]


class AsyncpgCursor(base.Cursor):
    def __init__(self, context, cursor):
        self._context = context
        self._cursor = cursor

    async def many(self, n, *, timeout=base.DEFAULT):
        if timeout is base.DEFAULT:
            timeout = self._context.timeout
        rows = await self._cursor.fetch(n, timeout=timeout)
        return self._context.process_rows(rows)

    async def next(self, *, timeout=base.DEFAULT):
        if timeout is base.DEFAULT:
            timeout = self._context.timeout
        row = await self._cursor.fetchrow(timeout=timeout)
        if not row:
            return None
        return self._context.process_rows([row])[0]

    async def forward(self, n, *, timeout=base.DEFAULT):
        if timeout is base.DEFAULT:
            timeout = self._context.timeout
        await self._cursor.forward(n, timeout=timeout)


class PreparedStatement(base.PreparedStatement):
    def __init__(self, prepared, clause=None):
        super().__init__(clause)
        self._prepared = prepared

    def _get_iterator(self, *params, **kwargs):
        return AsyncpgIterator(
            self.context, self._prepared.cursor(*params, **kwargs).__aiter__()
        )

    async def _get_cursor(self, *params, **kwargs):
        iterator = await self._prepared.cursor(*params, **kwargs)
        return AsyncpgCursor(self.context, iterator)

    async def _execute(self, params, one):
        if one:
            rv = await self._prepared.fetchrow(*params)
            if rv is None:
                rv = []
            else:
                rv = [rv]
        else:
            rv = await self._prepared.fetch(*params)
        return self._prepared.get_statusmsg(), rv


class DBAPICursor(base.DBAPICursor):
    def __init__(self, dbapi_conn):
        self._conn = dbapi_conn
        self._attributes = None
        self._status = None

    async def prepare(self, context, clause=None):
        timeout = context.timeout
        if timeout is None:
            conn = await self._conn.acquire(timeout=timeout)
        else:
            before = time.monotonic()
            conn = await self._conn.acquire(timeout=timeout)
            after = time.monotonic()
            timeout -= after - before
        prepared = await conn.prepare(context.statement, timeout=timeout)
        try:
            self._attributes = prepared.get_attributes()
        except TypeError:  # asyncpg <= 0.12.0
            self._attributes = []
        rv = PreparedStatement(prepared, clause)
        rv.context = context
        return rv

    async def async_execute(self, query, timeout, args, limit=0, many=False):
        if timeout is None:
            conn = await self._conn.acquire(timeout=timeout)
        else:
            before = time.monotonic()
            conn = await self._conn.acquire(timeout=timeout)
            after = time.monotonic()
            timeout -= after - before
        _protocol = getattr(conn, "_protocol")
        timeout = getattr(_protocol, "_get_timeout")(timeout)

        def executor(state, timeout_):
            if many:
                return _protocol.bind_execute_many(state, args, "", timeout_)
            else:
                return _protocol.bind_execute(state, args, "", limit, True, timeout_)

        with getattr(conn, "_stmt_exclusive_section"):
            result, stmt = await getattr(conn, "_do_execute")(query, executor, timeout)
            try:
                self._attributes = getattr(stmt, "_get_attributes")()
            except TypeError:  # asyncpg <= 0.12.0
                self._attributes = []
            if not many:
                result, self._status = result[:2]
            return result

    @property
    def description(self):
        return [((a[0], a[1][0]) + (None,) * 5) for a in self._attributes]

    def get_statusmsg(self):
        return self._status.decode()


class Pool(base.Pool):
    def __init__(self, url, loop, **kwargs):
        self._url = url
        self._loop = loop
        self._kwargs = kwargs
        self._pool = None

    async def _init(self):
        args = self._kwargs.copy()
        args.update(
            loop=self._loop,
            host=self._url.host,
            port=self._url.port,
            user=self._url.username,
            database=self._url.database,
            password=self._url.password,
        )
        self._pool = await asyncpg.create_pool(**args)
        return self

    def __await__(self):
        return self._init().__await__()

    @property
    def raw_pool(self):
        return self._pool

    async def acquire(self, *, timeout=None):
        return await self._pool.acquire(timeout=timeout)

    async def release(self, conn):
        await self._pool.release(conn)

    async def close(self):
        await self._pool.close()

    def repr(self, color):
        if color and not click:
            warnings.warn("Install click to get colorful repr.", ImportWarning)

        if color and click:
            # noinspection PyProtectedMember
            return "<{classname} max={max} min={min} cur={cur} use={use}>".format(
                classname=click.style(
                    self._pool.__class__.__module__
                    + "."
                    + self._pool.__class__.__name__,
                    fg="green",
                ),
                max=click.style(repr(self._pool._maxsize), fg="cyan"),
                min=click.style(repr(self._pool._minsize), fg="cyan"),
                cur=click.style(
                    repr(
                        len(
                            [
                                0
                                for con in self._pool._holders
                                if con._con and not con._con.is_closed()
                            ]
                        )
                    ),
                    fg="cyan",
                ),
                use=click.style(
                    repr(len([0 for con in self._pool._holders if con._in_use])),
                    fg="cyan",
                ),
            )
        else:
            # noinspection PyProtectedMember
            return "<{classname} max={max} min={min} cur={cur} use={use}>".format(
                classname=self._pool.__class__.__module__
                + "."
                + self._pool.__class__.__name__,
                max=self._pool._maxsize,
                min=self._pool._minsize,
                cur=len(
                    [
                        0
                        for con in self._pool._holders
                        if con._con and not con._con.is_closed()
                    ]
                ),
                use=len([0 for con in self._pool._holders if con._in_use]),
            )


class NullPool(base.Pool):
    # TODO: generic NullPool, abstracting connection part
    def __init__(self, url, loop, **kwargs):
        self._loop = loop
        self._kwargs = dict()
        for k in inspect.getfullargspec(asyncpg.connect).kwonlyargs:
            if k in kwargs:
                self._kwargs[k] = kwargs[k]
        self._kwargs.update(
            dict(
                host=url.host,
                port=url.port,
                user=url.username,
                database=url.database,
                password=url.password,
            )
        )
        self._count = 0

    def __await__(self):
        async def return_self():
            return self

        return return_self().__await__()

    @property
    def raw_pool(self):
        return self

    async def acquire(self, *, timeout=None):
        args = self._kwargs.copy()
        if timeout is not None:
            args.update(timeout=timeout)
        rv = await asyncpg.connect(loop=self._loop, **args)
        self._count += 1
        return rv

    async def release(self, conn):
        self._count -= 1
        await conn.close()

    async def close(self):
        pass

    def repr(self, color):
        if color and not click:
            warnings.warn("Install click to get colorful repr.", ImportWarning)

        if color and click:
            return "<{classname} max={max} min={min} cur={cur} use={cur}>".format(
                classname=click.style("NullPool", fg="green"),
                max=click.style("inf", fg="cyan"),
                min=click.style("0", fg="cyan"),
                cur=click.style(repr(self._count), fg="cyan"),
            )
        else:
            return "<NullPool max=inf min=0 cur={cur} use={cur}>".format(
                cur=self._count
            )


class Transaction(base.Transaction):
    def __init__(self, tx):
        self._tx = tx

    @property
    def raw_transaction(self):
        return self._tx

    async def begin(self):
        await self._tx.start()

    async def commit(self):
        await self._tx.commit()

    async def rollback(self):
        await self._tx.rollback()


class AsyncEnum(ENUM):
    async def create_async(self, bind=None, checkfirst=True):
        if not checkfirst or not await bind.dialect.has_type(
            bind, self.name, schema=self.schema
        ):
            await bind.status(CreateEnumType(self))

    async def drop_async(self, bind=None, checkfirst=True):
        if not checkfirst or await bind.dialect.has_type(
            bind, self.name, schema=self.schema
        ):
            await bind.status(DropEnumType(self))

    async def _on_table_create_async(self, target, bind, checkfirst=False, **kw):
        if (
            checkfirst
            or (not self.metadata and not kw.get("_is_metadata_operation", False))
            and not self._check_for_name_in_memos(checkfirst, kw)
        ):
            await self.create_async(bind=bind, checkfirst=checkfirst)

    async def _on_table_drop_async(self, target, bind, checkfirst=False, **kw):
        if (
            not self.metadata
            and not kw.get("_is_metadata_operation", False)
            and not self._check_for_name_in_memos(checkfirst, kw)
        ):
            await self.drop_async(bind=bind, checkfirst=checkfirst)

    async def _on_metadata_create_async(self, target, bind, checkfirst=False, **kw):
        if not self._check_for_name_in_memos(checkfirst, kw):
            await self.create_async(bind=bind, checkfirst=checkfirst)

    async def _on_metadata_drop_async(self, target, bind, checkfirst=False, **kw):
        if not self._check_for_name_in_memos(checkfirst, kw):
            await self.drop_async(bind=bind, checkfirst=checkfirst)


class GinoNullType(sqltypes.NullType):
    def result_processor(self, dialect, coltype):
        if coltype == JSON_COLTYPE:
            return JSON().result_processor(dialect, coltype)
        if coltype == JSONB_COLTYPE:
            return JSONB().result_processor(dialect, coltype)
        return super().result_processor(dialect, coltype)


class AsyncpgJSONPathType(json.JSONPathType):
    def bind_processor(self, dialect):
        super_proc = self.string_bind_processor(dialect)

        def process(value):
            assert isinstance(value, util.collections_abc.Sequence)
            if super_proc:
                return [super_proc(util.text_type(elem)) for elem in value]
            else:
                return [util.text_type(elem) for elem in value]

        return process


# noinspection PyAbstractClass
class AsyncpgDialect(PGDialect, base.AsyncDialectMixin):
    driver = "asyncpg"
    supports_native_decimal = True
    dbapi_class = AsyncpgDBAPI
    statement_compiler = AsyncpgCompiler
    execution_ctx_cls = AsyncpgExecutionContext
    cursor_cls = DBAPICursor
    init_kwargs = set(
        itertools.chain(
            *[
                inspect.getfullargspec(f).kwonlydefaults.keys()
                for f in [asyncpg.create_pool, asyncpg.connect]
            ]
        )
    )
    colspecs = util.update_copy(
        PGDialect.colspecs,
        {
            ENUM: AsyncEnum,
            sqltypes.Enum: AsyncEnum,
            sqltypes.NullType: GinoNullType,
            sqltypes.JSON.JSONPathType: AsyncpgJSONPathType,
        },
    )

    def __init__(self, *args, **kwargs):
        self._pool_kwargs = {}
        for k in self.init_kwargs:
            if k in kwargs:
                self._pool_kwargs[k] = kwargs.pop(k)
        super().__init__(*args, **kwargs)
        self._init_mixin()

    async def init_pool(self, url, loop, pool_class=None):
        if pool_class is None:
            pool_class = Pool
        return await pool_class(url, loop, init=self.on_connect(), **self._pool_kwargs)

    # noinspection PyMethodMayBeStatic
    def transaction(self, raw_conn, args, kwargs):
        return Transaction(raw_conn.transaction(*args, **kwargs))

    def on_connect(self):
        if self.isolation_level is not None:

            async def connect(conn):
                await self.set_isolation_level(conn, self.isolation_level)

            return connect
        else:
            return None

    async def set_isolation_level(self, connection, level):
        """
        Given an asyncpg connection, set its isolation level.

        """
        level = level.replace("_", " ")
        if level not in self._isolation_lookup:
            raise exc.ArgumentError(
                "Invalid value '%s' for isolation_level. "
                "Valid isolation levels for %s are %s"
                % (level, self.name, ", ".join(self._isolation_lookup))
            )
        await connection.execute(
            "SET SESSION CHARACTERISTICS AS TRANSACTION " "ISOLATION LEVEL %s" % level
        )
        await connection.execute("COMMIT")

    async def get_isolation_level(self, connection):
        """
        Given an asyncpg connection, return its isolation level.

        """
        val = await connection.fetchval("show transaction isolation level")
        return val.upper()

    async def has_schema(self, connection, schema):
        row = await connection.first(
            sql.text(
                "select nspname from pg_namespace " "where lower(nspname)=:schema"
            ).bindparams(
                sql.bindparam(
                    "schema", util.text_type(schema.lower()), type_=sqltypes.Unicode,
                )
            )
        )

        return bool(row)

    async def has_table(self, connection, table_name, schema=None):
        # seems like case gets folded in pg_class...
        if schema is None:
            row = await connection.first(
                sql.text(
                    "select relname from pg_class c join pg_namespace n on "
                    "n.oid=c.relnamespace where "
                    "pg_catalog.pg_table_is_visible(c.oid) "
                    "and relname=:name"
                ).bindparams(
                    sql.bindparam(
                        "name", util.text_type(table_name), type_=sqltypes.Unicode
                    ),
                )
            )
        else:
            row = await connection.first(
                sql.text(
                    "select relname from pg_class c join pg_namespace n on "
                    "n.oid=c.relnamespace where n.nspname=:schema and "
                    "relname=:name"
                ).bindparams(
                    sql.bindparam(
                        "name", util.text_type(table_name), type_=sqltypes.Unicode,
                    ),
                    sql.bindparam(
                        "schema", util.text_type(schema), type_=sqltypes.Unicode,
                    ),
                )
            )
        return bool(row)

    async def has_sequence(self, connection, sequence_name, schema=None):
        if schema is None:
            row = await connection.first(
                sql.text(
                    "SELECT relname FROM pg_class c join pg_namespace n on "
                    "n.oid=c.relnamespace where relkind='S' and "
                    "n.nspname=current_schema() "
                    "and relname=:name"
                ).bindparams(
                    sql.bindparam(
                        "name", util.text_type(sequence_name), type_=sqltypes.Unicode,
                    )
                )
            )
        else:
            row = await connection.first(
                sql.text(
                    "SELECT relname FROM pg_class c join pg_namespace n on "
                    "n.oid=c.relnamespace where relkind='S' and "
                    "n.nspname=:schema and relname=:name"
                ).bindparams(
                    sql.bindparam(
                        "name", util.text_type(sequence_name), type_=sqltypes.Unicode,
                    ),
                    sql.bindparam(
                        "schema", util.text_type(schema), type_=sqltypes.Unicode,
                    ),
                )
            )

        return bool(row)

    async def has_type(self, connection, type_name, schema=None):
        if schema is not None:
            query = """
            SELECT EXISTS (
                SELECT * FROM pg_catalog.pg_type t, pg_catalog.pg_namespace n
                WHERE t.typnamespace = n.oid
                AND t.typname = :typname
                AND n.nspname = :nspname
                )
                """
            query = sql.text(query)
        else:
            query = """
            SELECT EXISTS (
                SELECT * FROM pg_catalog.pg_type t
                WHERE t.typname = :typname
                AND pg_type_is_visible(t.oid)
                )
                """
            query = sql.text(query)
        query = query.bindparams(
            sql.bindparam(
                "typname", util.text_type(type_name), type_=sqltypes.Unicode,
            ),
        )
        if schema is not None:
            query = query.bindparams(
                sql.bindparam(
                    "nspname", util.text_type(schema), type_=sqltypes.Unicode,
                ),
            )
        return bool(await connection.scalar(query))



from .api import Gino  # NOQA
from .engine import GinoEngine, GinoConnection  # NOQA
from .exceptions import *  # NOQA
from .strategies import GinoStrategy  # NOQA


def create_engine(*args, **kwargs):
    """Shortcut for :func:`sqlalchemy.create_engine` with ``strategy="gino"``."""

    from sqlalchemy import create_engine

    kwargs.setdefault("strategy", "gino")
    return create_engine(*args, **kwargs)


def get_version():
    """Get current GINO version."""

    try:
        from importlib.metadata import version
    except ImportError:
        from importlib_metadata import version
    return version("gino")


# noinspection PyBroadException
try:
    __version__ = get_version()
except Exception:
    pass



class GinoException(Exception):
    pass


class NoSuchRowError(GinoException):
    pass


class UninitializedError(GinoException):
    pass


class UnknownJSONPropertyError(GinoException):
    pass


class MultipleResultsFound(GinoException):
    pass


class NoResultFound(GinoException):
    pass



import collections

import sqlalchemy as sa
from sqlalchemy.exc import InvalidRequestError

from . import json_support
from .exceptions import GinoException


class ColumnAttribute:
    """The type of the column wrapper attributes on GINO models.

    This is the core utility to enable GINO models so that:

    * Accessing a column attribute on a model class returns the column itself
    * Accessing a column attribute on a model instance returns the value for that column

    This utility is customizable by defining ``__attr_factory__`` in the model class.
    """

    def __init__(self, prop_name, column):
        self.prop_name = prop_name
        self.column = column

    def __get__(self, instance, owner):
        if instance is None:
            return self.column
        else:
            return instance.__values__.get(self.prop_name)

    def __set__(self, instance, value):
        instance.__values__[self.prop_name] = value

    def __delete__(self, instance):
        raise AttributeError("Cannot delete value.")


class InvertDict(dict):
    """A custom :class:`dict` that allows getting keys by values.

    Used internally by :class:`~Model`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inverted_dict = dict()
        for k, v in self.items():
            if v in self._inverted_dict:
                raise GinoException(
                    "Column name {} already maps to {}".format(
                        v, self._inverted_dict[v]
                    )
                )
            self._inverted_dict[v] = k

    def __setitem__(self, key, value):
        if value in self._inverted_dict and self._inverted_dict[value] != key:
            raise GinoException(
                "Column name {} already maps to {}".format(
                    value, self._inverted_dict[value]
                )
            )
        super().__setitem__(key, value)
        self._inverted_dict[value] = key

    def invert_get(self, value, default=None):
        """Get key by value.

        :param value: A value in this dict.
        :param default: If specified value doesn't exist, return default.
        :return: The corresponding key if the value is found, or default otherwise.
        """
        return self._inverted_dict.get(value, default)


class Dict(collections.OrderedDict):
    def __setitem__(self, key, value):
        if isinstance(value, sa.Column) and not value.name:
            value.name = value.key = key
        if isinstance(value, json_support.JSONProperty) and not value.name:
            value.name = key
        return super().__setitem__(key, value)


class ModelType(type):
    def _check_abstract(self):
        if self.__table__ is None:
            raise TypeError(
                "GINO model {} is abstract, no table is "
                "defined.".format(self.__name__)
            )

    def __iter__(self):
        self._check_abstract()
        # noinspection PyUnresolvedReferences
        return iter(self.__table__.columns)

    def __getattr__(self, item):
        try:
            if item in {"insert", "join", "outerjoin", "gino"}:
                self._check_abstract()
                return getattr(self.__table__, item)
            raise AttributeError
        except AttributeError:
            raise AttributeError(
                "type object '{}' has no attribute '{}'".format(self.__name__, item)
            )

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        return Dict()

    def __new__(mcs, name, bases, namespace, **kwargs):
        rv = type.__new__(mcs, name, bases, namespace)
        rv.__namespace__ = namespace
        if rv.__table__ is None:
            rv.__table__ = getattr(rv, "_init_table")(rv)
        return rv


def declared_attr(m):
    """
    Mark a class-level method as a factory of attribute.

    This is intended to be used as decorators on class-level methods of a
    :class:`~Model` class. When initializing the class as well as its
    subclasses, the decorated factory method will be called for each class, the
    returned result will be set on the class in place of the factory method
    under the same name.

    ``@declared_attr`` is implemented differently than
    :class:`~sqlalchemy.ext.declarative.declared_attr` of SQLAlchemy, but they
    are both more often used on mixins to dynamically declare indices or
    constraints (also works for column and ``__table_args__``, or even normal
    class attributes)::

        class TrackedMixin:
            created = db.Column(db.DateTime(timezone=True))

            @db.declared_attr
            def unique_id(cls):
                return db.Column(db.Integer())

            @db.declared_attr
            def unique_constraint(cls):
                return db.UniqueConstraint('unique_id')

            @db.declared_attr
            def poly(cls):
                if cls.__name__ == 'Thing':
                    return db.Column(db.Unicode())

            @db.declared_attr
            def __table_args__(cls):
                if cls.__name__ == 'Thing':
                    return db.UniqueConstraint('poly'),

    .. note::

        This doesn't work if the model already had a ``__table__``.

    """
    m.__declared_attr__ = True
    return m


class Model:
    """The base class of GINO models.

    This is not supposed to be sub-classed directly, :func:`~declarative_base` should
    be used instead to generate a base model class with a given
    :class:`~sqlalchemy.schema.MetaData`. By defining subclasses of a model, instances
    of :class:`sqlalchemy.schema.Table` will be created and added to the bound
    :class:`~sqlalchemy.schema.MetaData`. The columns of the
    :class:`~sqlalchemy.schema.Table` instance are defined as
    :class:`~sqlalchemy.schema.Column` attributes::

        from sqlalchemy import MetaData, Column, Integer, String
        from gino.declarative import declarative_base

        Model = declarative_base()

        class User(db.Model):
            __tablename__ = "users"
            id = Column(Integer(), primary_key=True)
            name = Column(String())

    The name of the columns are automatically set using the attribute name.

    An instance of a model will maintain a memory storage for values of all the defined
    column attributes. You can access these values by the same attribute name, or update
    with new values, just like normal Python objects::

        u = User()
        assert u.name is None

        u.name = "daisy"
        assert u.name == "daisy"

    .. note::

        Accessing column attributes on a model instance will NOT trigger any database
        operation.

    :class:`~sqlalchemy.schema.Constraint` and :class:`~sqlalchemy.schema.Index` are
    also allowed as model class attributes. Their attribute names are not used.

    A concrete model class can be used as a replacement of the
    :class:`~sqlalchemy.schema.Table` it reflects in SQLAlchemy queries. The model class
    is also iterable, yielding all the :class:`~sqlalchemy.schema.Column` instances
    defined in the model.

    Other built-in class attributes:

    * ``__metadata__``

      This is supposed to be set by :func:`~declarative_base` and used only during
      subclass construction. Still, this can be treated as a read-only attribute to
      find out which :class:`~sqlalchemy.schema.MetaData` this model is bound to.

    * ``__tablename__``

      This is a required attribute to define a concrete model, meaning a
      :class:`sqlalchemy.schema.Table` instance will be created, added to the bound
      :class:`~sqlalchemy.schema.MetaData` and set to the class attribute ``__table__``.
      Not defining ``__tablename__`` will result in an abstract model - no table
      instance will be created, and instances of an abstract model are meaningless.

    * ``__table__``

      This should usually be treated as an auto-generated read-only attribute storing
      the :class:`sqlalchemy.schema.Table` instance.

    * ``__attr_factory__``

      An attribute factory that is used to wrap the actual
      :class:`~sqlalchemy.schema.Column` instance on the model class, so that the access
      to the column attributes on model instances is redirected to the in-memory value
      store. The default factory is :class:`~ColumnAttribute`, can be override.

    * ``__values__``

      The internal in-memory value store as a :class:`dict`, only available on model
      instances. Accessing column attributes is equivalent to accessing ``__values__``.

    """

    __metadata__ = None
    __table__ = None
    __attr_factory__ = ColumnAttribute

    def __init__(self):
        self.__values__ = {}

    @classmethod
    def _init_table(cls, sub_cls):
        table_name = None
        columns = []
        inspected_args = []
        updates = {}
        column_name_map = InvertDict()
        visited = set()
        for each_cls in sub_cls.__mro__:
            for k, v in getattr(each_cls, "__namespace__", each_cls.__dict__).items():
                if k in visited:
                    continue
                visited.add(k)
                declared_callable_attr = callable(v) and getattr(
                    v, "__declared_attr__", False
                )
                if k == "__tablename__":
                    if declared_callable_attr:
                        table_name = v(sub_cls)
                    else:
                        table_name = v
                    continue
                if declared_callable_attr:
                    v = updates[k] = v(sub_cls)
                if isinstance(v, sa.Column):
                    v = v.copy()
                    if not v.name:
                        v.name = v.key = k
                    column_name_map[k] = v.name
                    columns.append(v)
                    updates[k] = sub_cls.__attr_factory__(k, v)
                elif isinstance(v, (sa.Index, sa.Constraint)):
                    inspected_args.append(v)
                elif isinstance(v, json_support.JSONProperty):
                    updates[k] = v
        if table_name is None:
            return
        sub_cls._column_name_map = column_name_map

        # handle __table_args__
        table_args = updates.get(
            "__table_args__", getattr(sub_cls, "__table_args__", None)
        )
        args, table_kw = (), {}
        if isinstance(table_args, dict):
            table_kw = table_args
        elif isinstance(table_args, tuple) and table_args:
            if isinstance(table_args[-1], dict):
                args, table_kw = table_args[0:-1], table_args[-1]
            else:
                args = table_args

        args = (*columns, *inspected_args, *args)
        for item in args:
            try:
                _table = getattr(item, "table", None)
            except InvalidRequestError:
                _table = None
            if _table is not None:
                raise ValueError(
                    "{} is already attached to another table. Please do not "
                    "use the same item twice. A common mistake is defining "
                    "constraints and indices in a super class - we are working"
                    " on making it possible."
                )
        rv = sa.Table(table_name, sub_cls.__metadata__, *args, **table_kw)
        for k, v in updates.items():
            setattr(sub_cls, k, v)

        json_prop_names = set()
        for each_cls in sub_cls.__mro__[::-1]:
            for k, v in each_cls.__dict__.items():
                if isinstance(v, json_support.JSONProperty):
                    if not v.name:
                        v.name = k
                    json_prop_names.add(v.prop_name)
                    json_col = getattr(
                        sub_cls.__dict__.get(v.prop_name), "column", None
                    )
                    if not isinstance(json_col, sa.Column) or not isinstance(
                        json_col.type, sa.JSON
                    ):
                        raise AttributeError(
                            '{} "{}" requires a JSON[B] column "{}" '
                            "which is not found or has a wrong type.".format(
                                type(v).__name__, v.name, v.prop_name,
                            )
                        )
        sub_cls.__json_prop_names__ = json_prop_names
        return rv


def declarative_base(metadata, model_classes=(Model,), name="Model"):
    """Create a base GINO model class for declarative table definition.

    :param metadata: A :class:`~sqlalchemy.schema.MetaData` instance to contain the
                     tables.
    :param model_classes: Base class(es) of the base model class to be created. Default:
                          :class:`~Model`.
    :param name: The class name of the base model class to be created. Default:
                 ``Model``.
    :return: A new base model class.
    """
    return ModelType(name, model_classes, {"__metadata__": metadata})


# noinspection PyProtectedMember
@sa.inspection._inspects(ModelType)
def inspect_model_type(target):
    target._check_abstract()
    return sa.inspection.inspect(target.__table__)


__all__ = [
    "ColumnAttribute",
    "Model",
    "declarative_base",
    "declared_attr",
    "InvertDict",
]



import weakref

import sqlalchemy as sa
from sqlalchemy.engine.url import make_url, URL
from sqlalchemy.sql.base import Executable
from sqlalchemy.sql.schema import SchemaItem

from .crud import CRUDModel
from .declarative import declarative_base, declared_attr
from .exceptions import UninitializedError
from .schema import GinoSchemaVisitor, patch_schema
from . import json_support


class GinoExecutor:
    """
    The default ``gino`` extension on
    :class:`~sqlalchemy.sql.expression.Executable` constructs for implicit
    execution.

    Instances of this class are created when visiting the ``gino`` property of
    :class:`~sqlalchemy.sql.expression.Executable` instances (also referred as
    queries or clause elements), for example::

        await User.query.gino.first()

    This allows GINO to add the asynchronous query APIs (:meth:`all`,
    :meth:`first`, :meth:`one`, :meth:`one_or_none`, :meth:`scalar`,
    :meth:`status`, :meth:`iterate`) to SQLAlchemy query clauses without
    messing up with existing synchronous ones.
    Calling these asynchronous query APIs has the same restriction - the
    relevant metadata (the :class:`Gino` instance) must be bound to an engine,
    or an :exc:`AttributeError` will be raised.

    .. note::

        Executable clause elements that are completely irrelevant with any
        table - for example ``db.select([db.text('now()')])`` - has no
        metadata, hence no engine. Therefore, this will always fail::

            await db.select([db.text('now()')]).gino.scalar()

        You should use :meth:`conn.scalar() <.engine.GinoConnection.scalar>`,
        :meth:`engine.scalar() <.engine.GinoEngine.scalar>` or even
        :meth:`db.scalar() <.Gino.scalar>` in this case.

    """

    __slots__ = ("_query",)

    def __init__(self, query):
        self._query = query

    @property
    def query(self):
        """
        Get back the chained :class:`~sqlalchemy.sql.expression.Executable`.

        In a chained query calls, occasionally the previous query clause is
        needed after a ``.gino.`` chain, you can use ``.query.`` to resume the
        chain back. For example::

            await User.query.gino.model(FOUser).query.where(...).gino.all()

        """
        return self._query

    def model(self, model):
        """
        Shortcut to set execution option ``model`` in a chaining call.

        Read :meth:`~gino.engine.GinoConnection.execution_options` for more
        information.

        """
        if model is not None:
            model = weakref.ref(model)
        self._query = self._query.execution_options(model=model)
        return self

    def return_model(self, switch):
        """
        Shortcut to set execution option ``return_model`` in a chaining call.

        Read :meth:`~gino.engine.GinoConnection.execution_options` for more
        information.

        """
        self._query = self._query.execution_options(return_model=switch)
        return self

    def timeout(self, timeout):
        """
        Shortcut to set execution option ``timeout`` in a chaining call.

        Read :meth:`~gino.engine.GinoConnection.execution_options` for more
        information.

        """
        self._query = self._query.execution_options(timeout=timeout)
        return self

    def load(self, value):
        """
        Shortcut to set execution option ``loader`` in a chaining call.

        For example to load ``Book`` instances with their authors::

            query = Book.join(User).select()
            books = await query.gino.load(Book.load(author=User)).all()

        Read :meth:`~gino.engine.GinoConnection.execution_options` for more
        information.

        """
        self._query = self._query.execution_options(loader=value)
        return self

    async def all(self, *multiparams, **params):
        """
        Returns :meth:`engine.all() <.engine.GinoEngine.all>` with this query
        as the first argument, and other arguments followed, where ``engine``
        is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        return await self._query.bind.all(self._query, *multiparams, **params)

    async def first(self, *multiparams, **params):
        """
        Returns :meth:`engine.first() <.engine.GinoEngine.first>` with this
        query as the first argument, and other arguments followed, where
        ``engine`` is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        return await self._query.bind.first(self._query, *multiparams, **params)

    async def one_or_none(self, *multiparams, **params):
        """
        Returns :meth:`engine.one_or_none() <.engine.GinoEngine.one_or_none>`
        with this query as the first argument, and other arguments followed,
        where ``engine`` is the :class:`~.engine.GinoEngine` to which the
        metadata (:class:`Gino`) is bound, while metadata is found in this
        query.

        """
        return await self._query.bind.one_or_none(self._query, *multiparams, **params)

    async def one(self, *multiparams, **params):
        """
        Returns :meth:`engine.one() <.engine.GinoEngine.one>` with this query
        as the first argument, and other arguments followed, where ``engine``
        is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        return await self._query.bind.one(self._query, *multiparams, **params)

    async def scalar(self, *multiparams, **params):
        """
        Returns :meth:`engine.scalar() <.engine.GinoEngine.scalar>` with this
        query as the first argument, and other arguments followed, where
        ``engine`` is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        return await self._query.bind.scalar(self._query, *multiparams, **params)

    async def status(self, *multiparams, **params):
        """
        Returns :meth:`engine.status() <.engine.GinoEngine.status>` with this
        query as the first argument, and other arguments followed, where
        ``engine`` is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        return await self._query.bind.status(self._query, *multiparams, **params)

    def iterate(self, *multiparams, **params):
        """
        Returns :meth:`engine.iterate() <.engine.GinoEngine.iterate>` with this
        query as the first argument, and other arguments followed, where
        ``engine`` is the :class:`~.engine.GinoEngine` to which the metadata
        (:class:`Gino`) is bound, while metadata is found in this query.

        """
        connection = self._query.bind.current_connection
        if connection is None:
            raise ValueError("No Connection in context, please provide one")
        return connection.iterate(self._query, *multiparams, **params)


class _BindContext:
    def __init__(self, *args):
        self._args = args

    async def __aenter__(self):
        api, bind, loop, kwargs = self._args
        return await api.set_bind(bind, loop, **kwargs)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._args[0].pop_bind().close()


class Gino(sa.MetaData):
    """
    All-in-one API class of GINO, providing several shortcuts.

    This class is a subclass of SQLAlchemy
    :class:`~sqlalchemy.schema.MetaData`, therefore its instances can be used
    as a normal :class:`~sqlalchemy.schema.MetaData` object, e.g. used in
    `Alembic <http://alembic.zzzcomputing.com/>`_. In usual cases, you would
    want to define one global :class:`~.Gino` instance, usually under the name
    of ``db``, representing the database used in your application.

    You may define tables in `the official way <http://bit.ly/2G25fdc>`_
    SQLAlchemy core recommended, but more often in GINO we define model classes
    with ``db.Model`` as their parent class to represent tables, for its
    objective interface and CRUD operations. Please read :doc:`/how-to/crud`
    for more information.

    For convenience, :class:`Gino` instance delegated all properties publicly
    exposed by :mod:`sqlalchemy`, so that you can define tables / models
    without importing :mod:`sqlalchemy`::

        id = db.Column(db.BigInteger(), primary_key=True)

    Similar to :class:`~sqlalchemy.schema.MetaData`, a :class:`~.Gino` object
    can bind to a :class:`~gino.engine.GinoEngine` instance, hereby allowing
    `"implicit execution" <http://bit.ly/2oTUcKY>`_ through the ``gino``
    extension on :class:`~sqlalchemy.sql.expression.Executable` or
    :class:`~sqlalchemy.schema.SchemaItem` constructs::

        await User.query.gino.first()
        await db.gino.create_all()

    Differently, GINO encourages the use of implicit execution and manages
    transactional context correctly.

    Binding to a connection object is not supported.

    To set a bind property, you can simply set your
    :class:`~gino.engine.GinoEngine` object on :attr:`db.bind <Gino.bind>`, or
    set it to ``None`` to unbind. However, the creation of engine usually
    happens at the same time. Therefore, GINO provided several convenient ways
    doing so:

    1. :meth:`~Gino.with_bind` returning an asynchronous context manager::

        async with db.with_bind('postgresql://...') as engine:

    2. :meth:`~Gino.set_bind` and :meth:`~Gino.pop_bind`::

        engine = await db.set_bind('postgresql://...')
        await db.pop_bind().close()

    3. Directly ``await`` on :class:`~.Gino` instance::

        db = await gino.Gino('postgresql://...')
        await db.pop_bind().close()

    .. note::

        SQLAlchemy allows creating the engine by::

            metadata.bind = 'postgresql://...'

        While in GINO this only sets a string to :attr:`~.Gino.bind`, because
        creating an engine requires ``await``, which is exactly what
        :meth:`~.Gino.set_bind` does.

    At last, :class:`Gino` delegates all query APIs on the bound
    :class:`~.engine.GinoEngine`.

    """

    model_base_classes = (CRUDModel,)
    """
    Overridable default model classes to build the :attr:`Model`.

    Default is :class:`(CRUDModel, ) <gino.crud.CRUDModel>`.

    """

    query_executor = GinoExecutor
    """
    The overridable ``gino`` extension class on
    :class:`~sqlalchemy.sql.expression.Executable`.

    This class will be set as the getter method of the property ``gino`` on
    :class:`~sqlalchemy.sql.expression.Executable` and its subclasses, if
    ``ext`` and ``query_ext`` arguments are both ``True``. Default is
    :class:`GinoExecutor`.

    """

    schema_visitor = GinoSchemaVisitor
    """
    The overridable ``gino`` extension class on
    :class:`~sqlalchemy.schema.SchemaItem`.

    This class will be set as the getter method of the property ``gino`` on
    :class:`~sqlalchemy.schema.SchemaItem` and its subclasses, if ``ext`` and
    ``schema_ext`` arguments are both ``True``. Default is
    :class:`~gino.schema.GinoSchemaVisitor`.

    """

    no_delegate = {"create_engine", "engine_from_config"}
    """
    A set of symbols from :mod:`sqlalchemy` which is not delegated by
    :class:`Gino`.

    """

    def __init__(
        self,
        bind=None,
        model_classes=None,
        query_ext=True,
        schema_ext=True,
        ext=True,
        **kwargs
    ):
        """
        :param bind: A :class:`~.engine.GinoEngine` instance to bind. Also
                     accepts string or :class:`~sqlalchemy.engine.url.URL`,
                     which will be passed to
                     :func:`~gino.create_engine` when this :class:`Gino`
                     instance is awaited. Default is ``None``.
        :param model_classes: A :class:`tuple` of base class and mixin classes
                              to create the :attr:`~.Gino.Model` class. Default
                              is :class:`(CRUDModel, ) <gino.crud.CRUDModel>`.
        :param query_ext: Boolean value to control the installation of the
                          ``gino`` extension on
                          :class:`~sqlalchemy.sql.expression.Executable` for
                          implicit execution. Default is to install (``True``).
        :param schema_ext: Boolean value to control the installation of the
                           ``gino`` extension on
                           :class:`~sqlalchemy.schema.SchemaItem` for implicit
                           execution. Default is to install (``True``).
        :param ext: Boolean value to control the installation of the two
                    ``gino`` extensions. ``False`` for no extension at all,
                    while it depends on the two individual switches when this
                    is set to ``True`` (default).
        :param kwargs: Other arguments accepted by
                       :class:`~sqlalchemy.schema.MetaData`.

        """
        super().__init__(bind=bind, **kwargs)
        if model_classes is None:
            model_classes = self.model_base_classes
        self._model = declarative_base(self, model_classes)
        self.declared_attr = declared_attr
        self.quoted_name = sa.sql.quoted_name
        for mod in json_support, sa:
            for key in mod.__all__:
                if not hasattr(self, key) and key not in self.no_delegate:
                    setattr(self, key, getattr(mod, key))
        if ext:
            if query_ext:
                Executable.gino = property(self.query_executor)
            if schema_ext:
                SchemaItem.gino = property(self.schema_visitor)
                patch_schema(self)

    # noinspection PyPep8Naming
    @property
    def Model(self):
        """
        Declarative base class for models, subclass of
        :class:`gino.declarative.Model`. Defining subclasses of this class will
        result new tables added to this :class:`Gino` metadata.

        """
        return self._model

    @property
    def bind(self):
        """
        An :class:`~.engine.GinoEngine` to which this :class:`Gino` is bound.

        This is a simple property with no getter or setter hook - what you set
        is what you get. To achieve the same result as it is in SQLAlchemy -
        setting a string or :class:`~sqlalchemy.engine.url.URL` and getting an
        engine instance, use :meth:`set_bind` (or ``await`` on this
        :class:`Gino` object after setting a string or
        :class:`~sqlalchemy.engine.url.URL`).

        """
        if self._bind is None:
            return _PlaceHolder(UninitializedError("Gino engine is not initialized."))
        return self._bind

    # noinspection PyMethodOverriding,PyAttributeOutsideInit
    @bind.setter
    def bind(self, bind):
        self._bind = bind

    async def set_bind(self, bind, loop=None, **kwargs):
        """
        Bind self to the given :class:`~.engine.GinoEngine` and return it.

        If the given ``bind`` is a string or
        :class:`~sqlalchemy.engine.url.URL`, all arguments will be sent to
        :func:`~gino.create_engine` to create a new engine, and return it.

        :return: :class:`~.engine.GinoEngine`

        """
        if isinstance(bind, str):
            bind = make_url(bind)
        if isinstance(bind, URL):
            from . import create_engine

            bind = await create_engine(bind, loop=loop, **kwargs)
        self.bind = bind
        return bind

    def pop_bind(self):
        """
        Unbind self, and return the bound engine.

        This is usually used in a chained call to close the engine::

            await db.pop_bind().close()

        :return: :class:`~.engine.GinoEngine` or ``None`` if self is not bound.

        """
        bind, self.bind = self.bind, None
        return bind

    def with_bind(self, bind, loop=None, **kwargs):
        """
        Shortcut for :meth:`set_bind` and :meth:`pop_bind` plus closing engine.

        This method accepts the same arguments of
        :func:`~gino.create_engine`. This allows inline creating an engine and
        binding self on enter, and unbinding self and closing the engine on
        exit::

            async with db.with_bind('postgresql://...') as engine:
                # play with engine

        :return: An asynchronous context manager.

        """
        return _BindContext(self, bind, loop, kwargs)

    def __await__(self):
        async def init():
            await self.set_bind(self.bind)
            return self

        return init().__await__()

    def compile(self, elem, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.compile()
        <.engine.GinoEngine.compile>`.

        """
        return self.bind.compile(elem, *multiparams, **params)

    async def all(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.all() <.engine.GinoEngine.all>`.

        """
        return await self.bind.all(clause, *multiparams, **params)

    async def first(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.first() <.engine.GinoEngine.first>`.

        """
        return await self.bind.first(clause, *multiparams, **params)

    async def one_or_none(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.one_or_none()
        <.engine.GinoEngine.one_or_none>`.

        """
        return await self.bind.one_or_none(clause, *multiparams, **params)

    async def one(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.one() <.engine.GinoEngine.first>`.

        """
        return await self.bind.one(clause, *multiparams, **params)

    async def scalar(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.scalar() <.engine.GinoEngine.scalar>`.

        """
        return await self.bind.scalar(clause, *multiparams, **params)

    async def status(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.status() <.engine.GinoEngine.status>`.

        """
        return await self.bind.status(clause, *multiparams, **params)

    def iterate(self, clause, *multiparams, **params):
        """
        A delegate of :meth:`GinoEngine.iterate()
        <.engine.GinoEngine.iterate>`.

        """
        return self.bind.iterate(clause, *multiparams, **params)

    def acquire(self, *args, **kwargs):
        """
        A delegate of :meth:`GinoEngine.acquire()
        <.engine.GinoEngine.acquire>`.

        """
        return self.bind.acquire(*args, **kwargs)

    def transaction(self, *args, **kwargs):
        """
        A delegate of :meth:`GinoEngine.transaction()
        <.engine.GinoEngine.transaction>`.

        """
        return self.bind.transaction(*args, **kwargs)


class _PlaceHolder:
    __slots__ = "_exception"

    def __init__(self, exception):
        self._exception = exception

    def __getattribute__(self, item):
        if item == "_exception":
            return super().__getattribute__(item)
        raise self._exception

    def __setattr__(self, key, value):
        if key == "_exception":
            return super().__setattr__(key, value)
        raise self._exception



import asyncio
import weakref

from sqlalchemy import util

# noinspection PyProtectedMember
from ..engine import _SAConnection, _SAEngine, _DBAPIConnection
from ..loader import Loader

DEFAULT = object()


class BaseDBAPI:
    paramstyle = "numeric"
    Error = Exception

    # noinspection PyPep8Naming
    @staticmethod
    def Binary(x):
        return x


class DBAPICursor:
    def execute(self, statement, parameters):
        pass

    def executemany(self, statement, parameters):
        pass

    @property
    def description(self):
        raise NotImplementedError

    async def prepare(self, context, clause=None):
        raise NotImplementedError

    async def async_execute(self, query, timeout, args, limit=0, many=False):
        raise NotImplementedError

    def get_statusmsg(self):
        raise NotImplementedError


class Pool:
    @property
    def raw_pool(self):
        raise NotImplementedError

    async def acquire(self, *, timeout=None):
        raise NotImplementedError

    async def release(self, conn):
        raise NotImplementedError

    async def close(self):
        raise NotImplementedError

    def repr(self, color):
        return repr(self)


class Transaction:
    @property
    def raw_transaction(self):
        raise NotImplementedError

    async def begin(self):
        raise NotImplementedError

    async def commit(self):
        raise NotImplementedError

    async def rollback(self):
        raise NotImplementedError


class PreparedStatement:
    def __init__(self, clause=None):
        self.context = None
        self.clause = clause

    def iterate(self, *params, **kwargs):
        return _PreparedIterableCursor(self, params, kwargs)

    async def _do_execute(
        self, multiparams, params, one=False, return_model=True, status=False
    ):
        ctx = self.context.connection.execute(
            self.clause, *multiparams, **params
        ).context
        if ctx.executemany:
            raise ValueError(
                "PreparedStatement does not support multiple " "parameters."
            )
        if ctx.statement != self.context.statement:
            raise AssertionError(
                "Prepared statement generated different SQL with parameters"
            )
        params = []
        for val in ctx.parameters[0]:
            params.append(val)
        msg, rows = await self._execute(params, one)
        if status:
            return msg
        item = self.context.process_rows(rows, return_model=return_model)
        if one:
            if item:
                item = item[0]
            else:
                item = None
        return item

    async def all(self, *multiparams, **params):
        return await self._do_execute(multiparams, params)

    async def first(self, *multiparams, **params):
        return await self._do_execute(multiparams, params, one=True)

    async def scalar(self, *multiparams, **params):
        rv = await self._do_execute(multiparams, params, one=True, return_model=False)
        if rv:
            return rv[0]
        else:
            return None

    async def status(self, *multiparams, **params):
        return await self._do_execute(multiparams, params, status=True)

    def _get_iterator(self, *params, **kwargs):
        raise NotImplementedError

    async def _get_cursor(self, *params, **kwargs):
        raise NotImplementedError

    async def _execute(self, params, one):
        raise NotImplementedError


class _PreparedIterableCursor:
    def __init__(self, prepared, params, kwargs):
        self._prepared = prepared
        self._params = params
        self._kwargs = kwargs

    def __aiter__(self):
        return getattr(self._prepared, "_get_iterator")(*self._params, **self._kwargs)

    def __await__(self):
        return getattr(self._prepared, "_get_cursor")(
            *self._params, **self._kwargs
        ).__await__()


class _IterableCursor:
    def __init__(self, context):
        self._context = context

    async def _iterate(self):
        prepared = await self._context.cursor.prepare(self._context)
        return prepared.iterate(
            *self._context.parameters[0], timeout=self._context.timeout
        )

    async def _get_cursor(self):
        return await (await self._iterate())

    def __aiter__(self):
        return _LazyIterator(self._iterate)

    def __await__(self):
        return self._get_cursor().__await__()


class _LazyIterator:
    def __init__(self, init):
        self._init = init
        self._iter = None

    async def __anext__(self):
        if self._iter is None:
            self._iter = (await self._init()).__aiter__()
        return await self._iter.__anext__()


class _ResultProxy:
    _metadata = True

    def __init__(self, context):
        self._context = context

    @property
    def context(self):
        return self._context

    async def execute(self, one=False, return_model=True, status=False):
        context = self._context

        param_groups = []
        for params in context.parameters:
            replace_params = []
            for val in params:
                if asyncio.iscoroutine(val):
                    val = await val
                replace_params.append(val)
            param_groups.append(replace_params)

        cursor = context.cursor
        if context.executemany:
            return await cursor.async_execute(
                context.statement, context.timeout, param_groups, many=True
            )
        else:
            args = param_groups[0]
            rows = await cursor.async_execute(
                context.statement, context.timeout, args, 1 if one else 0
            )
            item = context.process_rows(rows, return_model=return_model)
            if one:
                if item:
                    item = item[0]
                else:
                    item = None
            if status:
                item = cursor.get_statusmsg(), item
            return item

    def iterate(self):
        if self._context.executemany:
            raise ValueError("too many multiparams")
        return _IterableCursor(self._context)

    async def prepare(self, clause):
        return await self._context.cursor.prepare(self._context, clause)

    def _soft_close(self):
        pass


class Cursor:
    async def many(self, n, *, timeout=DEFAULT):
        raise NotImplementedError

    async def next(self, *, timeout=DEFAULT):
        raise NotImplementedError

    async def forward(self, n, *, timeout=DEFAULT):
        raise NotImplementedError


class ExecutionContextOverride:
    def _compiled_first_opt(self, key, default=DEFAULT):
        rv = DEFAULT
        opts = getattr(getattr(self, "compiled", None), "execution_options", None)
        if opts:
            rv = opts.get(key, DEFAULT)
        if rv is DEFAULT:
            # noinspection PyUnresolvedReferences
            rv = self.execution_options.get(key, default)
        if rv is DEFAULT:
            raise LookupError("No such execution option!")
        return rv

    @util.memoized_property
    def return_model(self):
        return self._compiled_first_opt("return_model", True)

    @util.memoized_property
    def model(self):
        rv = self._compiled_first_opt("model", None)
        if isinstance(rv, weakref.ref):
            rv = rv()
        return rv

    @util.memoized_property
    def timeout(self):
        return self._compiled_first_opt("timeout", None)

    @util.memoized_property
    def loader(self):
        return self._compiled_first_opt("loader", None)

    def process_rows(self, rows, return_model=True):
        # noinspection PyUnresolvedReferences
        rv = rows = super().get_result_proxy().process_rows(rows)
        loader = self.loader
        if loader is None and self.model is not None:
            loader = Loader.get(self.model)
        if loader is not None and return_model and self.return_model:
            ctx = {}
            rv = []
            loader = Loader.get(loader)
            for row in rows:
                obj, distinct = loader.do_load(row, ctx)
                if distinct:
                    rv.append(obj)
        return rv

    def get_result_proxy(self):
        return _ResultProxy(self)

    @classmethod
    def _init_compiled_prepared(
        cls, dialect, connection, dbapi_connection, compiled, parameters
    ):
        self = cls.__new__(cls)
        self.root_connection = connection
        self._dbapi_connection = dbapi_connection
        self.dialect = connection.dialect

        self.compiled = compiled

        # this should be caught in the engine before
        # we get here
        assert compiled.can_execute

        self.execution_options = compiled.execution_options.union(
            connection._execution_options
        )

        self.result_column_struct = (
            compiled._result_columns,
            compiled._ordered_columns,
            compiled._textual_ordered_columns,
        )

        self.unicode_statement = util.text_type(compiled)
        if not dialect.supports_unicode_statements:
            self.statement = self.unicode_statement.encode(self.dialect.encoding)
        else:
            self.statement = self.unicode_statement

        self.isinsert = compiled.isinsert
        self.isupdate = compiled.isupdate
        self.isdelete = compiled.isdelete
        self.is_text = compiled.isplaintext

        self.executemany = False

        self.cursor = self.create_cursor()

        if self.isinsert or self.isupdate or self.isdelete:
            self.is_crud = True
            self._is_explicit_returning = bool(compiled.statement._returning)
            self._is_implicit_returning = bool(
                compiled.returning and not compiled.statement._returning
            )

        if self.dialect.positional:
            self.parameters = [dialect.execute_sequence_format()]
        else:
            self.parameters = [{}]
        self.compiled_parameters = [{}]

        return self

    @classmethod
    def _init_statement_prepared(
        cls, dialect, connection, dbapi_connection, statement, parameters
    ):
        """Initialize execution context for a string SQL statement."""

        self = cls.__new__(cls)
        self.root_connection = connection
        self._dbapi_connection = dbapi_connection
        self.dialect = connection.dialect
        self.is_text = True

        # plain text statement
        self.execution_options = connection._execution_options

        if self.dialect.positional:
            self.parameters = [dialect.execute_sequence_format()]
        else:
            self.parameters = [{}]

        self.executemany = False

        if not dialect.supports_unicode_statements and isinstance(
            statement, util.text_type
        ):
            self.unicode_statement = statement
            self.statement = dialect._encoder(statement)[0]
        else:
            self.statement = self.unicode_statement = statement

        self.cursor = self.create_cursor()
        return self


class AsyncDialectMixin:
    cursor_cls = DBAPICursor
    dbapi_class = BaseDBAPI

    def _init_mixin(self):
        self._sa_conn = _SAConnection(
            _SAEngine(self), _DBAPIConnection(self.cursor_cls)
        )

    @classmethod
    def dbapi(cls):
        return cls.dbapi_class

    def compile(self, elem, *multiparams, **params):
        context = self._sa_conn.execute(elem, *multiparams, **params).context
        if context.executemany:
            return context.statement, context.parameters
        else:
            return context.statement, context.parameters[0]

    async def init_pool(self, url, loop):
        raise NotImplementedError

    def transaction(self, raw_conn, args, kwargs):
        raise NotImplementedError



"""Generate typing stubs for extensions.

    $ python -m gino.ext

"""
import sys
import os

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


if __name__ == "__main__":
    base_dir = os.path.dirname(os.path.abspath(__file__))
    cmd = sys.argv[1] if len(sys.argv) == 2 else ""
    eps = list(entry_points().get("gino.extensions", []))

    if cmd == "stub":
        added = False
        for ep in eps:
            path = os.path.join(base_dir, ep.name + ".pyi")
            if not os.path.exists(path):
                added = True
                print("Adding " + path)
                with open(path, "w") as f:
                    f.write("from " + ep.value + " import *")
        if not added:
            print("Stub files are up to date.")

    elif cmd == "clean":
        removed = False
        for filename in os.listdir(base_dir):
            if filename.endswith(".pyi"):
                removed = True
                path = os.path.join(base_dir, filename)
                print("Removing " + path)
                os.remove(path)
        if not removed:
            print("No stub files found.")

    elif cmd == "list":
        name_size = max(len(ep.name) for ep in eps)
        value_size = max(len(ep.value) for ep in eps)
        for ep in eps:
            path = os.path.join(base_dir, ep.name + ".pyi")
            if not os.path.exists(path):
                path = "no stub file"
            print(
                "%s -> gino.ext.%s (%s)"
                % (ep.value.ljust(value_size), ep.name.ljust(name_size), path)
            )

    else:
        print("Manages GINO extensions:")
        print()
        print("  python -m gino.ext COMMAND")
        print()
        print("Available commands:")
        print()
        print("  stub    Generate gino/ext/*.pyi stub files for type checking.")
        print("  clean   Remove the generated stub files.")
        print("  list    List installed GINO extensions.")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gino', 'gino.dialects', 'gino.ext']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.2.16,<1.4', 'asyncpg>=0.18,<1.0']

extras_require = \
{':python_version < "3.7"': ['contextvars>=2.4,<3.0'],
 ':python_version < "3.8"': ['importlib_metadata>=1.3.0,<2.0.0'],
 'aiohttp:python_version >= "3.5.3" and python_version < "4.0.0"': ['gino-aiohttp>=0.1.0,<0.2.0'],
 'quart:python_version >= "3.7" and python_version < "4.0"': ['gino-quart>=0.1.0,<0.2.0'],
 'sanic:python_version >= "3.6" and python_version < "4.0"': ['gino-sanic>=0.1.0,<0.2.0'],
 'starlette:python_version >= "3.6" and python_version < "4.0"': ['gino-starlette>=0.1.1,<0.2.0'],
 'tornado:python_version >= "3.5.2" and python_version < "4.0.0"': ['gino-tornado>=0.1.0,<0.2.0']}

entry_points = \
{'sqlalchemy.dialects': ['asyncpg = gino.dialects.asyncpg:AsyncpgDialect',
                         'postgresql.asyncpg = '
                         'gino.dialects.asyncpg:AsyncpgDialect']}

setup_kwargs = {
    'name': 'gino',
    'version': '1.0.1',
    'description': 'GINO Is Not ORM - a Python asyncio ORM on SQLAlchemy core.',
    'long_description': '======\n|GINO|\n======\n\n.. image:: https://img.shields.io/pypi/v/gino?logo=python&logoColor=white\n        :alt: PyPI Release Version\n        :target: https://pypi.python.org/pypi/gino\n\n.. image:: https://img.shields.io/github/workflow/status/python-gino/gino/test?label=test&logo=github\n        :alt: GitHub Workflow Status for tests\n        :target: https://github.com/python-gino/gino/actions?query=workflow%3Atest\n\n.. image:: https://img.shields.io/github/workflow/status/python-gino/gino/docs?label=docs&logo=github\n        :alt: GitHub Workflow Status for docs\n        :target: https://python-gino.org/docs/\n\n.. image:: https://img.shields.io/codacy/coverage/b6a59cdf5ca64eab9104928d4f9bbb97?logo=codacy\n        :alt: Codacy coverage\n        :target: https://app.codacy.com/gh/python-gino/gino/dashboard\n\n.. image:: https://img.shields.io/badge/Dependabot-active-brightgreen?logo=dependabot\n        :target: https://app.dependabot.com/accounts/python-gino/projects/129260\n        :alt: Dependabot\n\n.. image:: https://img.shields.io/gitter/room/python-gino/Lobby?logo=gitter\n        :target: https://gitter.im/python-gino/Lobby\n        :alt: Gitter chat\n\n\nGINO - GINO Is Not ORM - is a lightweight asynchronous ORM built on top of\nSQLAlchemy_ core for Python asyncio_. GINO 1.0 supports only PostgreSQL_ with asyncpg_.\n\n* Free software: BSD license\n* Requires: Python 3.5\n* GINO is developed proudly with |PyCharm|.\n\n\nHome\n----\n\n`python-gino.org <https://python-gino.org/>`__\n\n\nDocumentation\n-------------\n\n* English_\n* Chinese_\n\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install gino\n\n\nFeatures\n--------\n\n* Robust SQLAlchemy-asyncpg bi-translator with no hard hack\n* Asynchronous SQLAlchemy-alike engine and connection\n* Asynchronous dialect API\n* Asynchronous-friendly CRUD objective models\n* Well-considered contextual connection and transaction management\n* Reusing native SQLAlchemy core to build queries with grammar sugars\n* Support SQLAlchemy ecosystem, e.g. Alembic_ for migration\n* `Community support <https://github.com/python-gino/>`_ for Starlette_/FastAPI_, aiohttp_, Sanic_, Tornado_ and Quart_\n* Rich PostgreSQL JSONB support\n\n\n.. _SQLAlchemy: https://www.sqlalchemy.org/\n.. _asyncpg: https://github.com/MagicStack/asyncpg\n.. _PostgreSQL: https://www.postgresql.org/\n.. _asyncio: https://docs.python.org/3/library/asyncio.html\n.. _Alembic: https://bitbucket.org/zzzeek/alembic\n.. _Sanic: https://github.com/channelcat/sanic\n.. _Tornado: http://www.tornadoweb.org/\n.. _Quart: https://gitlab.com/pgjones/quart/\n.. _English: https://python-gino.org/docs/en/\n.. _Chinese: https://python-gino.org/docs/zh/\n.. _aiohttp: https://github.com/aio-libs/aiohttp\n.. _Starlette: https://www.starlette.io/\n.. _FastAPI: https://fastapi.tiangolo.com/\n.. |PyCharm| image:: ./docs/images/pycharm.svg\n        :height: 20px\n        :target: https://www.jetbrains.com/?from=GINO\n\n.. |GINO| image:: ./docs/theme/static/logo.svg\n        :alt: GINO\n        :height: 64px\n        :target: https://python-gino.org/\n',
    'author': 'Fantix King',
    'author_email': 'fantix.king@gmail.com',
    'maintainer': 'Tony Wang',
    'maintainer_email': 'wwwjfy@gmail.com',
    'url': 'https://python-gino.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import asyncio
import collections
import functools
import sys
import time
from contextvars import ContextVar

from sqlalchemy.engine import Engine, Connection
from sqlalchemy.sql import schema

from .aiocontextvars import patch_asyncio
from .exceptions import MultipleResultsFound, NoResultFound
from .transaction import GinoTransaction

patch_asyncio()


class _BaseDBAPIConnection:
    _reset_agent = None
    gino_conn = None

    def __init__(self, cursor_cls):
        self._cursor_cls = cursor_cls
        self._closed = False

    def commit(self):
        pass

    def cursor(self):
        return self._cursor_cls(self)

    @property
    def raw_connection(self):
        raise NotImplementedError

    async def acquire(self, *, timeout=None):
        if self._closed:
            raise ValueError("This connection is already released permanently.")
        return await self._acquire(timeout)

    async def _acquire(self, timeout):
        raise NotImplementedError

    async def release(self, permanent):
        if permanent:
            self._closed = True
        return await self._release()

    async def _release(self):
        raise NotImplementedError


class _DBAPIConnection(_BaseDBAPIConnection):
    def __init__(self, cursor_cls, pool=None):
        super().__init__(cursor_cls)
        self._pool = pool
        self._conn = None
        self._lock = asyncio.Lock()

    @property
    def raw_connection(self):
        return self._conn

    async def _acquire(self, timeout):
        try:
            if timeout is None:
                await self._lock.acquire()
            else:
                before = time.monotonic()
                await asyncio.wait_for(self._lock.acquire(), timeout=timeout)
                after = time.monotonic()
                timeout -= after - before
            if self._conn is None:
                self._conn = await self._pool.acquire(timeout=timeout)
            return self._conn
        finally:
            self._lock.release()

    async def _release(self):
        conn, self._conn = self._conn, None
        if conn is None:
            return False
        await self._pool.release(conn)
        return True


class _ReusingDBAPIConnection(_BaseDBAPIConnection):
    def __init__(self, cursor_cls, root):
        super().__init__(cursor_cls)
        self._root = root

    @property
    def raw_connection(self):
        return self._root.raw_connection

    async def _acquire(self, timeout):
        return await self._root.acquire(timeout=timeout)

    async def _release(self):
        pass


# noinspection PyPep8Naming,PyMethodMayBeStatic
class _bypass_no_param:
    def keys(self):
        return []


_bypass_no_param = _bypass_no_param()


# noinspection PyAbstractClass
class _SAConnection(Connection):
    def _execute_context(self, dialect, constructor, statement, parameters, *args):
        if parameters == [_bypass_no_param]:
            constructor = getattr(
                self.dialect.execution_ctx_cls,
                constructor.__name__ + "_prepared",
                constructor,
            )
        return super()._execute_context(
            dialect, constructor, statement, parameters, *args
        )


# noinspection PyAbstractClass
class _SAEngine(Engine):
    _connection_cls = _SAConnection

    def __init__(self, dialect, **kwargs):
        super().__init__(None, dialect, None, **kwargs)


class _AcquireContext:
    __slots__ = ["_acquire", "_conn"]

    def __init__(self, acquire):
        self._acquire = acquire
        self._conn = None

    async def __aenter__(self):
        self._conn = await self._acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        conn, self._conn = self._conn, None
        await conn.release()

    def __await__(self):
        return self._acquire().__await__()


class _TransactionContext:
    __slots__ = ["_conn_ctx", "_tx_ctx"]

    def __init__(self, conn_ctx, args):
        self._conn_ctx = conn_ctx
        self._tx_ctx = args

    async def __aenter__(self):
        conn = await self._conn_ctx.__aenter__()
        try:
            args, kwargs = self._tx_ctx
            self._tx_ctx = conn.transaction(*args, **kwargs)
            return await self._tx_ctx.__aenter__()
        except Exception:
            await self._conn_ctx.__aexit__(*sys.exc_info())
            raise

    async def __aexit__(self, *exc_info):
        try:
            tx, self._tx_ctx = self._tx_ctx, None
            return await tx.__aexit__(*exc_info)
        except Exception:
            exc_info = sys.exc_info()
            raise
        finally:
            await self._conn_ctx.__aexit__(*exc_info)


class GinoConnection:
    """
    Represents an actual database connection.

    This is the root of all query API like :meth:`all`, :meth:`first`,
    :meth:`one`, :meth:`one_or_none`, :meth:`scalar` or :meth:`status`,
    those on engine or query are simply wrappers of methods in this class.

    Usually instances of this class are created by :meth:`.GinoEngine.acquire`.

    .. note::

        :class:`.GinoConnection` may refer to zero or one underlying database
        connection - when a :class:`.GinoConnection` is acquired with
        ``lazy=True``, the underlying connection may still be in the pool,
        until a query API is called or :meth:`get_raw_connection` is called.

        Oppositely, one underlying database connection can be shared by many
        :class:`.GinoConnection` instances when they are acquired with
        ``reuse=True``. The actual database connection is only returned to the
        pool when the **root** :class:`.GinoConnection` is released. Read more
        in :meth:`GinoEngine.acquire` method.

    .. seealso::

        :doc:`/explanation/engine`

    """

    # noinspection PyProtectedMember
    schema_for_object = schema._schema_getter(None)
    """A SQLAlchemy compatibility attribute, don't use it for now, it bites."""

    def __init__(self, dialect, sa_conn, stack=None):
        self._dialect = dialect
        self._sa_conn = sa_conn
        self._stack = stack

    @property
    def _dbapi_conn(self):
        return self._sa_conn.connection

    @property
    def raw_connection(self):
        """
        The current underlying database connection instance, type depends on
        the dialect in use. May be ``None`` if self is a lazy connection.

        """
        return self._dbapi_conn.raw_connection

    async def get_raw_connection(self, *, timeout=None):
        """
        Get the underlying database connection, acquire one if none present.

        :param timeout: Seconds to wait for the underlying acquiring
        :return: Underlying database connection instance depending on the
                 dialect in use
        :raises: :class:`~asyncio.TimeoutError` if the acquiring timed out

        """
        return await self._dbapi_conn.acquire(timeout=timeout)

    async def release(self, *, permanent=True):
        """
        Returns the underlying database connection to its pool.

        If ``permanent=False``, this connection will be set in lazy mode with
        underlying database connection returned, the next query on this
        connection will cause a new database connection acquired. This is
        useful when this connection may still be useful again later, while some
        long-running I/O operations are about to take place, which should not
        take up one database connection or even transaction for that long time.

        Otherwise with ``permanent=True`` (default), this connection will be
        marked as closed after returning to pool, and be no longer usable
        again.

        If this connection is a reusing connection, then only this connection
        is closed (depending on ``permanent``), the reused underlying
        connection will **not** be returned back to the pool.

        Practically it is recommended to return connections in the reversed
        order as they are borrowed, but if this connection is a reused
        connection with still other opening connections reusing it, then on
        release the underlying connection **will be** returned to the pool,
        with all the reusing connections losing an available underlying
        connection. The availability of further operations on those reusing
        connections depends on the given ``permanent`` value.

        .. seealso::

            :meth:`.GinoEngine.acquire`

        """
        if permanent and self._stack is not None:
            dbapi_conn = self._stack.remove(lambda x: x.gino_conn is self)
            if dbapi_conn:
                await dbapi_conn.release(True)
            else:
                raise ValueError("This connection is already released.")
        else:
            await self._dbapi_conn.release(permanent)

    @property
    def dialect(self):
        """
        The :class:`~sqlalchemy.engine.interfaces.Dialect` in use, inherited
        from the engine created this connection.

        """
        return self._dialect

    def _execute(self, clause, multiparams, params):
        return self._sa_conn.execute(clause, *multiparams, **params)

    async def all(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns all results as a list.

        This method accepts the same parameters taken by SQLAlchemy
        :meth:`~sqlalchemy.engine.Connectable.execute`. You can pass in a raw
        SQL string, or *any* SQLAlchemy query clauses.

        If the given query clause is built by CRUD models, then the returning
        rows will be turned into relevant model objects (Only one type of model
        per query is supported for now, no relationship support yet). See
        :meth:`execution_options` for more information.

        If the given parameters are parsed as "executemany" - bulk inserting
        multiple rows in one call for example, the returning result from
        database will be discarded and this method will return ``None``.

        """
        result = self._execute(clause, multiparams, params)
        return await result.execute()

    async def first(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns the first result.

        If the query returns no result, this method will return ``None``.

        See :meth:`all` for common query comments.

        """
        result = self._execute(clause, multiparams, params)
        return await result.execute(one=True)

    async def one_or_none(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns at most one result.

        If the query returns no result, this method will return ``None``.
        If the query returns multiple results, this method will raise
        :class:`~gino.exceptions.MultipleResultsFound`.

        See :meth:`all` for common query comments.

        """
        result = self._execute(clause, multiparams, params)
        ret = await result.execute()

        if ret is None or len(ret) == 0:
            return None

        if len(ret) == 1:
            return ret[0]

        raise MultipleResultsFound("Multiple rows found for one_or_none()")

    async def one(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns exactly one result.

        If the query returns no result, this method will raise
        :class:`~gino.exceptions.NoResultFound`.
        If the query returns multiple results, this method will raise
        :class:`~gino.exceptions.MultipleResultsFound`.

        See :meth:`all` for common query comments.

        """
        try:
            ret = await self.one_or_none(clause, *multiparams, **params)
        except MultipleResultsFound:
            raise MultipleResultsFound("Multiple rows found for one()")

        if ret is None:
            raise NoResultFound("No row was found for one()")

        return ret

    async def scalar(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns the first result.

        If the query returns no result, this method will return ``None``.

        See :meth:`all` for common query comments.

        """
        result = self._execute(clause, multiparams, params)
        rv = await result.execute(one=True, return_model=False)
        if rv:
            return rv[0]
        else:
            return None

    async def status(self, clause, *multiparams, **params):
        """
        Runs the given query in database, returns the query status.

        The returning query status depends on underlying database and the
        dialect in use. For asyncpg it is a string, you can parse it like this:
        https://git.io/v7oze

        """
        result = self._execute(clause, multiparams, params)
        return await result.execute(status=True)

    def transaction(self, *args, **kwargs):
        """
        Starts a database transaction.

        There are two ways using this method: **managed** as an asynchronous
        context manager::

            async with conn.transaction() as tx:
                # run query in transaction

        or **manually** awaited::

            tx = await conn.transaction()
            try:
                # run query in transaction
                await tx.commit()
            except Exception:
                await tx.rollback()
                raise

        Where the ``tx`` is an instance of the
        :class:`~gino.transaction.GinoTransaction` class, feel free to read
        more about it.

        In the first managed mode, the transaction is automatically committed
        on exiting the context block, or rolled back if an exception was raised
        which led to the exit of the context. In the second manual mode, you'll
        need to manually call the
        :meth:`~gino.transaction.GinoTransaction.commit` or
        :meth:`~gino.transaction.GinoTransaction.rollback` methods on need.

        If this is a lazy connection, entering a transaction will cause a new
        database connection acquired if none was present.

        Transactions may support nesting depending on the dialect in use. For
        example in asyncpg, starting a second transaction on the same
        connection will create a save point in the database.

        For now, the parameters are directly passed to underlying database
        driver, read :meth:`asyncpg.connection.Connection.transaction` for
        asyncpg.

        """
        return GinoTransaction(self, args, kwargs)

    def iterate(self, clause, *multiparams, **params):
        """
        Creates a server-side cursor in database for large query results.

        Cursors must work within transactions::

            async with conn.transaction():
                async for user in conn.iterate(User.query):
                    # handle each user without loading all users into memory

        Alternatively, you can manually control how the cursor works::

            async with conn.transaction():
                cursor = await conn.iterate(User.query)
                user = await cursor.next()
                users = await cursor.many(10)

        Read more about how :class:`~gino.dialects.base.Cursor` works.

        Similarly, this method takes the same parameters as :meth:`all`.

        """
        result = self._execute(clause, multiparams, params)
        return result.iterate()

    def execution_options(self, **opt):
        """
        Set non-SQL options for the connection which take effect during
        execution.

        This method returns a copy of this :class:`.GinoConnection` which
        references the same underlying database connection, but with the given
        execution options set on the copy. Therefore, it is a good practice to
        discard the copy immediately after use, for example::

            row = await conn.execution_options(model=None).first(User.query)

        This is very much the same as SQLAlchemy
        :meth:`~sqlalchemy.engine.base.Connection.execution_options`, it
        actually does pass the execution options to the underlying SQLAlchemy
        :class:`~sqlalchemy.engine.base.Connection`. Furthermore, GINO added a
        few execution options:

        :param return_model: Boolean to control whether the returning results
          should be loaded into model instances, where the model class is
          defined in another execution option ``model``. Default is ``True``.

        :param model: Specifies the type of model instance to create on return.
          This has no effect if ``return_model`` is set to ``False``. Usually
          in queries built by CRUD models, this execution option is
          automatically set. For now, GINO only supports loading each row into
          one type of model object, relationships are not supported. Please use
          multiple queries for that. ``None`` for no postprocessing (default).

        :param timeout: Seconds to wait for the query to finish. ``None`` for
          no time out (default).

        :param loader: A loader expression to load the database rows into
          specified objective structure. It can be either:

          * A model class, so that the query will yield model instances of this
            class. It is your responsibility to make sure all the columns of
            this model is selected in the query.
          * A :class:`~sqlalchemy.schema.Column` instance, so that each result
            will be only a single value of this column. Please note, if you
            want to achieve fetching the very first value, you should use
            :meth:`~gino.engine.GinoConnection.first` instead of
            :meth:`~gino.engine.GinoConnection.scalar`. However, using directly
            :meth:`~gino.engine.GinoConnection.scalar` is a more direct way.
          * A tuple nesting more loader expressions recursively.
          * A :func:`callable` function that will be called for each row to
            fully customize the result. Two positional arguments will be passed
            to the function: the first is the :class:`row
            <sqlalchemy.engine.RowProxy>` instance, the second is a context
            object which is only present if nested else ``None``.
          * A :class:`~gino.loader.Loader` instance directly.
          * Anything else will be treated as literal values thus returned as
            whatever they are.

        """
        return type(self)(self._dialect, self._sa_conn.execution_options(**opt))

    async def _run_visitor(self, visitorcallable, element, **kwargs):
        await visitorcallable(self.dialect, self, **kwargs).traverse_single(element)

    async def prepare(self, clause):
        return await self._execute(clause, (_bypass_no_param,), {}).prepare(clause)


class _ContextualStack:
    __slots__ = ("_ctx", "_stack")

    def __init__(self, ctx):
        self._ctx = ctx
        self._stack = ctx.get()
        if self._stack is None:
            self._stack = collections.deque()
            ctx.set(self._stack)

    def __bool__(self):
        return bool(self._stack)

    @property
    def top(self):
        return self._stack[-1]

    def push(self, value):
        self._stack.append(value)

    def remove(self, checker):
        for i in range(len(self._stack)):
            if checker(self._stack[-1]):
                rv = self._stack.pop()
                if self._stack:
                    self._stack.rotate(-i)
                else:
                    self._ctx.set(None)
                return rv
            else:
                self._stack.rotate(1)


class GinoEngine:
    """
    Connects a :class:`~.dialects.base.Pool` and
    :class:`~sqlalchemy.engine.interfaces.Dialect` together to provide a source
    of database connectivity and behavior.

    A :class:`.GinoEngine` object is instantiated publicly using the
    :func:`gino.create_engine` function or
    :func:`db.set_bind() <gino.api.Gino.set_bind>` method.

    .. seealso::

        :doc:`/explanation/engine`

    """

    connection_cls = GinoConnection
    """Customizes the connection class to use, default is
    :class:`.GinoConnection`."""

    def __init__(
        self, dialect, pool, loop, logging_name=None, echo=None, execution_options=None
    ):
        self._sa_engine = _SAEngine(
            dialect,
            logging_name=logging_name,
            echo=echo,
            execution_options=execution_options,
        )
        self._dialect = dialect
        self._pool = pool
        self._loop = loop
        self._ctx = ContextVar("gino", default=None)

    @property
    def dialect(self):
        """
        Read-only property for the
        :class:`~sqlalchemy.engine.interfaces.Dialect` of this engine.

        """
        return self._dialect

    @property
    def raw_pool(self):
        """
        Read-only access to the underlying database connection pool instance.
        This depends on the actual dialect in use, :class:`~asyncpg.pool.Pool`
        of asyncpg for example.

        """
        return self._pool.raw_pool

    def acquire(self, *, timeout=None, reuse=False, lazy=False, reusable=True):
        """
        Acquire a connection from the pool.

        There are two ways using this method - as an asynchronous context
        manager::

            async with engine.acquire() as conn:
                # play with the connection

        which will guarantee the connection is returned to the pool when
        leaving the ``async with`` block; or as a coroutine::

            conn = await engine.acquire()
            try:
                # play with the connection
            finally:
                await conn.release()

        where the connection should be manually returned to the pool with
        :meth:`conn.release() <.GinoConnection.release>`.

        Within the same context (usually the same :class:`~asyncio.Task`, see
        also :doc:`/how-to/transaction`), a nesting acquire by default re

        :param timeout: Block up to ``timeout`` seconds until there is one free
          connection in the pool. Default is ``None`` - block forever until
          succeeded. This has no effect when ``lazy=True``, and depends on the
          actual situation when ``reuse=True``.

        :param reuse: Reuse the latest reusable acquired connection (before
          it's returned to the pool) in current context if there is one, or
          borrow a new one if none present. Default is ``False`` for always
          borrow a new one. This is useful when you are in a nested method call
          series, wishing to use the same connection without passing it around
          as parameters. See also: :doc:`/how-to/transaction`. A reusing
          connection is not reusable even if ``reusable=True``. If the reused
          connection happened to be a lazy one, then the reusing connection is
          lazy too.

        :param lazy: Don't acquire the actual underlying connection yet - do it
          only when needed. Default is ``False`` for always do it immediately.
          This is useful before entering a code block which may or may not make
          use of a given connection object. Feeding in a lazy connection will
          save the borrow-return job if the connection is never used. If
          setting ``reuse=True`` at the same time, then the reused connection -
          if any - applies the same laziness. For example, reusing a lazy
          connection with ``lazy=False`` will cause the reused connection to
          acquire an underlying connection immediately.

        :param reusable: Mark this connection as reusable or otherwise. This
          has no effect if it is a reusing connection. All reusable connections
          are placed in a stack, any reusing acquire operation will always
          reuse the top (latest) reusable connection. One reusable connection
          may be reused by several reusing connections - they all share one
          same underlying connection. Acquiring a connection with
          ``reusable=False`` and ``reusing=False`` makes it a cleanly isolated
          connection which is only referenced once here.

        :return: A :class:`.GinoConnection` object.

        """
        return _AcquireContext(
            functools.partial(self._acquire, timeout, reuse, lazy, reusable)
        )

    async def _acquire(self, timeout, reuse, lazy, reusable):
        stack = _ContextualStack(self._ctx)
        if reuse and stack:
            dbapi_conn = _ReusingDBAPIConnection(self._dialect.cursor_cls, stack.top)
            reusable = False
        else:
            dbapi_conn = _DBAPIConnection(self._dialect.cursor_cls, self._pool)
        rv = self.connection_cls(
            self._dialect,
            _SAConnection(self._sa_engine, dbapi_conn),
            stack if reusable else None,
        )
        dbapi_conn.gino_conn = rv
        if not lazy:
            await dbapi_conn.acquire(timeout=timeout)
        if reusable:
            stack.push(dbapi_conn)
        return rv

    @property
    def current_connection(self):
        """
        Gets the most recently acquired reusable connection in the context.
        ``None`` if there is no such connection.

        :return: :class:`.GinoConnection`

        """
        stack = self._ctx.get()
        if stack:
            return stack[-1].gino_conn

    async def close(self):
        """
        Close the engine, by closing the underlying pool.

        """
        await self._pool.close()

    async def all(self, clause, *multiparams, **params):
        """
        Acquires a connection with ``reuse=True`` and runs
        :meth:`~.GinoConnection.all` on it. ``reuse=True`` means you can safely
        do this without borrowing more than one underlying connection::

            async with engine.acquire():
                await engine.all('SELECT ...')

        The same applies for other query methods.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.all(clause, *multiparams, **params)

    async def first(self, clause, *multiparams, **params):
        """
        Runs :meth:`~.GinoConnection.first`, See :meth:`.all`.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.first(clause, *multiparams, **params)

    async def one_or_none(self, clause, *multiparams, **params):
        """
        Runs :meth:`~.GinoConnection.one_or_none`, See :meth:`.all`.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.one_or_none(clause, *multiparams, **params)

    async def one(self, clause, *multiparams, **params):
        """
        Runs :meth:`~.GinoConnection.one`, See :meth:`.all`.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.one(clause, *multiparams, **params)

    async def scalar(self, clause, *multiparams, **params):
        """
        Runs :meth:`~.GinoConnection.scalar`, See :meth:`.all`.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.scalar(clause, *multiparams, **params)

    async def status(self, clause, *multiparams, **params):
        """
        Runs :meth:`~.GinoConnection.status`. See also :meth:`.all`.

        """
        async with self.acquire(reuse=True) as conn:
            return await conn.status(clause, *multiparams, **params)

    def compile(self, clause, *multiparams, **params):
        """
        A shortcut for :meth:`~gino.dialects.base.AsyncDialectMixin.compile` on
        the dialect, returns raw SQL string and parameters according to the
        rules of the dialect.

        """
        return self._dialect.compile(clause, *multiparams, **params)

    def transaction(self, *args, timeout=None, reuse=True, reusable=True, **kwargs):
        """
        Borrows a new connection and starts a transaction with it.

        Different to :meth:`.GinoConnection.transaction`, transaction on engine
        level supports only managed usage::

            async with engine.transaction() as tx:
                # play with transaction here

        Where the implicitly acquired connection is available as
        :attr:`tx.connection <gino.transaction.GinoTransaction.connection>`.

        By default, :meth:`.transaction` acquires connection with
        ``reuse=True`` and ``reusable=True``, that means it by default tries to
        create a nested transaction instead of a new transaction on a new
        connection. You can change the default behavior by setting these two
        arguments.

        The other arguments are the same as
        :meth:`~.GinoConnection.transaction` on connection.

        .. seealso::

            :meth:`.GinoEngine.acquire`

            :meth:`.GinoConnection.transaction`

            :class:`~gino.transaction.GinoTransaction`

        :return: A asynchronous context manager that yields a
          :class:`~gino.transaction.GinoTransaction`

        """
        return _TransactionContext(
            self.acquire(timeout=timeout, reuse=reuse, reusable=reusable),
            (args, kwargs),
        )

    def iterate(self, clause, *multiparams, **params):
        """
        Creates a server-side cursor in database for large query results.

        This requires that there is a reusable connection in the current
        context, and an active transaction is present. Then its
        :meth:`.GinoConnection.iterate` is executed and returned.

        """
        connection = self.current_connection
        if connection is None:
            raise ValueError("No Connection in context, please provide one")
        return connection.iterate(clause, *multiparams, **params)

    def update_execution_options(self, **opt):
        """Update the default execution_options dictionary
        of this :class:`.GinoEngine`.

        .. seealso::

            :meth:`sqlalchemy.engine.Engine.update_execution_options`

            :meth:`.GinoConnection.execution_options`

        """
        self._sa_engine.update_execution_options(**opt)

    async def _run_visitor(self, *args, **kwargs):
        async with self.acquire(reuse=True) as conn:
            await getattr(conn, "_run_visitor")(*args, **kwargs)

    def repr(self, color=False):
        return self._pool.repr(color)

    def __repr__(self):
        return self.repr()
